import asyncio
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies.uis.tui.tui import TUI


class Job:
    def __init__(self, name, **kwargs):
        self.name = name
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return f'<{type(self).__name__} {self.name!r}>'


class JobInfo:
    def __init__(self, *, job=None, widget=None, container=None):
        self.job = job
        self.widget = widget
        self.container = container

    def __repr__(self):
        return f'<{type(self).__name__} {self.job.name!r}>'


@pytest.fixture
def tui(event_loop):
    tui = TUI()
    assert tui._loop is event_loop
    return tui


def test_uncaught_exception(tui, mocker):
    mocker.patch.object(tui, '_exit')

    counter = 0

    def bad_callback(task):
        nonlocal counter
        counter += 1
        raise RuntimeError(f'bad, bad, bad: {counter}')

    async def with_running_loop():
        for _ in range(3):
            task = asyncio.create_task(AsyncMock()())
            task.add_done_callback(bad_callback)
            await task

    tui._loop.run_until_complete(with_running_loop())

    assert type(tui._exception) is RuntimeError
    assert str(tui._exception) == 'bad, bad, bad: 1'
    assert tui._exit.call_args_list == [call(), call(), call()]


def test_add_jobs(tui, mocker):
    jobs = {
        'a': Job(name='a'),
        'b': Job(name='b'),
        'c': Job(name='c'),
    }

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tui, '_add_job'), '_add_job')
    mocks.attach_mock(mocker.patch.object(tui, '_update_jobs_container'), '_update_jobs_container')
    mocks.attach_mock(mocker.patch.object(tui, '_connect_jobs'), '_connect_jobs')

    tui.add_jobs(*jobs.values())

    assert mocks.mock_calls == [
        call._add_job(jobs['a']),
        call._add_job(jobs['b']),
        call._add_job(jobs['c']),
        call._update_jobs_container(),
        call._connect_jobs((jobs['a'], jobs['b'], jobs['c'])),
    ]


def test__add_job_detects_job_name_duplicate(tui, mocker):
    mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    mocker.patch('upsies.uis.tui.tui.to_container')

    tui._add_job(Job(name='a'))
    tui._add_job(Job(name='b'))
    with pytest.raises(RuntimeError, match=r'^Conflicting job name: b$'):
        tui._add_job(Job(name='b'))


def test__add_job_gracefully_ignores_adding_exact_job_duplicate(tui, mocker):
    mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    mocker.patch('upsies.uis.tui.tui.to_container')

    jobs = {
        'a': Job(name='a'),
        'b': Job(name='b'),
        'c': Job(name='c'),
        'd': Job(name='d'),
    }

    tui._add_job(jobs['a'])
    tui._add_job(jobs['b'])
    tui._add_job(jobs['c'])
    tui._add_job(jobs['b'])
    tui._add_job(jobs['d'])
    assert list(tui._jobs) == list(jobs)
    assert [jobinfo.job for jobinfo in tui._jobs.values()] == list(jobs.values())


def test__add_job_creates_JobWidget_and_Container(tui, mocker):
    JobWidget_mock = mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    to_container_mock = mocker.patch('upsies.uis.tui.tui.to_container')
    job = Job(name='a')
    tui._add_job(job)
    assert tuple(tui._jobs) == (job.name,)
    assert tui._jobs[job.name].widget == JobWidget_mock.return_value
    assert tui._jobs[job.name].container == to_container_mock.return_value
    assert JobWidget_mock.call_args_list == [call(job, tui._app)]
    assert to_container_mock.call_args_list == [call(JobWidget_mock.return_value)]


def test__update_jobs_container(tui, mocker):
    tui._jobs = {
        # Interactive jobs
        'ai': JobInfo(job=Job('a', is_enabled=False, was_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(id='aw')),
        'bi': JobInfo(job=Job('b', is_enabled=False, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(id='bw')),
        'ci': JobInfo(job=Job('c', is_enabled=True, was_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(id='cw')),
        'di': JobInfo(job=Job('d', is_enabled=True, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(id='dw')),
        'ei': JobInfo(job=Job('e', is_enabled=True, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(id='ew')),

        # Background jobs
        '1b': JobInfo(job=Job('1', is_enabled=False, was_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(id='1w')),
        '2b': JobInfo(job=Job('2', is_enabled=False, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(id='2w')),
        '3b': JobInfo(job=Job('3', is_enabled=True, was_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(id='3w')),
        '4b': JobInfo(job=Job('4', is_enabled=True, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(id='4w')),
        '5b': JobInfo(job=Job('5', is_enabled=True, was_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(id='5w')),
    }
    jobs_container_id = id(tui._jobs_container)

    # tui._layout.focus() may raise ValueError, which should be ignored.
    def layout_focus(container):
        if container in (
                tui._jobs['bi'].container,
                tui._jobs['ei'].container,
                tui._jobs['2b'].container,
                tui._jobs['4b'].container,
        ):
            raise ValueError(f'Unfocusable container: {container!r}')

    mocker.patch.object(tui._layout, 'focus', side_effect=layout_focus)

    def assert_jobs_container(*children, focused):
        tui._update_jobs_container()

        exp_children = [tui._jobs[child].container for child in children]
        assert tui._jobs_container.children == exp_children
        assert id(tui._jobs_container) == jobs_container_id

        if focused:
            focused_jobinfo = tui._jobs[focused]
            assert tui._focused_jobinfo is focused_jobinfo
            assert tui._layout.focus.call_args_list == [call(focused_jobinfo.container)]
        else:
            assert tui._focused_jobinfo is None
            assert tui._layout.focus.call_args_list == []
        tui._layout.focus.reset_mock()

    # First interactive, started and unfinished job is focused.
    # Enabled background jobs are listed.
    assert_jobs_container('di', '3b', '4b', '5b', focused='di')

    # Finishing background job should have no effect
    tui._jobs['2b'].job.is_enabled = True
    assert_jobs_container('di', '2b', '3b', '4b', '5b', focused='di')

    # Finishing focused job should focus next interactive job.
    tui._jobs['di'].job.is_finished = True
    assert_jobs_container('di', 'ei', '2b', '3b', '4b', '5b', focused='ei')

    # Finishing focused job should unfocus if there is no interactive, started
    # and unfinished jobs.
    tui._jobs['ei'].job.is_finished = True
    assert_jobs_container('di', 'ei', '2b', '3b', '4b', '5b', focused=None)

    # Enabling another interactive, started and unfinished job should display it
    # in order of the job list.
    tui._jobs['bi'].job.is_enabled = True
    assert_jobs_container('bi', 'di', 'ei', '2b', '3b', '4b', '5b', focused='bi')

    # Enabling another interactive, started and unfinished job should have no
    # effect until the focused job finishes. (Focus should not be stolen.)
    tui._jobs['ai'].job.is_enabled = True
    tui._jobs['ai'].job.was_started = True
    assert_jobs_container('bi', 'di', 'ei', '2b', '3b', '4b', '5b', focused='bi')

    # As soon as the focused job finishes, the previously enabled job may now
    # grab focus.
    tui._jobs['bi'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '2b', '3b', '4b', '5b', focused='ai')

    # Start and finish all remaining background jobs.
    tui._jobs['1b'].job.is_enabled = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')
    tui._jobs['5b'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')
    tui._jobs['3b'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')
    tui._jobs['1b'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')
    tui._jobs['2b'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')
    tui._jobs['4b'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused='ai')

    # The last interactive job finishes and focus should be null.
    tui._jobs['ai'].job.is_finished = True
    assert_jobs_container('ai', 'bi', 'di', 'ei', '1b', '2b', '3b', '4b', '5b', focused=None)


def test__connect_jobs(tui):
    mocks = Mock()
    jobs = [
        mocks.foo,
        mocks.bar,
        mocks.baz,
    ]
    tui._connect_jobs(jobs)
    assert mocks.mock_calls == [
        call.foo.signal.register('finished', tui._handle_job_finished),
        call.foo.signal.register('refresh_job_list', tui._refresh_jobs),
        call.bar.signal.register('finished', tui._handle_job_finished),
        call.bar.signal.register('refresh_job_list', tui._refresh_jobs),
        call.baz.signal.register('finished', tui._handle_job_finished),
        call.baz.signal.register('refresh_job_list', tui._refresh_jobs),
    ]


@pytest.mark.parametrize(
    argnames='finished_job, enabled_jobs, exp_mock_calls',
    argvalues=(
        pytest.param(
            Job(name='x', exit_code=1, is_finished=True),
            [
                Job(name='foo', is_finished=False),
            ],
            [call._refresh_jobs(), call._exit()],
            id='Finished job failed',
        ),
        pytest.param(
            Job(name='x', exit_code=0, is_finished=True),
            [
                Job(name='foo', is_finished=False),
            ],
            [call._refresh_jobs()],
            id='Finished job succeeded, only other job is not finished',
        ),
        pytest.param(
            Job(name='x', exit_code=0, is_finished=True),
            [
                Job(name='foo', is_finished=True),
            ],
            [call._refresh_jobs(), call._exit()],
            id='Finished job succeeded, only other job is finished',
        ),
        pytest.param(
            Job(name='x', exit_code=0, is_finished=True),
            [
                Job(name='foo', is_finished=True),
                Job(name='bar', is_finished=True),
                Job(name='baz', is_finished=True),
            ],
            [call._refresh_jobs(), call._exit()],
            id='Finished job succeeded, all other jobs are finished',
        ),
        pytest.param(
            Job(name='x', exit_code=0, is_finished=True),
            [
                Job(name='foo', is_finished=True),
                Job(name='bar', is_finished=False),
                Job(name='baz', is_finished=True),
            ],
            [call._refresh_jobs()],
            id='Finished job succeeded, one other job is not finished',
        ),
    ),
)
def test__handle_job_finished(finished_job, enabled_jobs, exp_mock_calls, tui, mocker):
    mocks = Mock()
    mocker.patch.object(type(tui), '_enabled_jobs', PropertyMock(return_value=[
        Mock(job=job) for job in enabled_jobs
    ]))
    mocks.attach_mock(mocker.patch.object(tui, '_exit'), '_exit')
    mocks.attach_mock(mocker.patch.object(tui, '_refresh_jobs'), '_refresh_jobs')

    tui._handle_job_finished(finished_job)

    assert mocks.mock_calls == exp_mock_calls


def test__refresh_jobs(tui, mocker):
    enabled_jobs = [
        Job(name='foo'),
        Job(name='bar'),
        Job(name='baz'),
    ]

    mocks = Mock()
    mocker.patch.object(type(tui), '_enabled_jobs', PropertyMock(return_value=[
        JobInfo(job=job) for job in enabled_jobs
    ]))
    mocks.attach_mock(mocker.patch.object(tui, '_start_enabled_jobs'), '_start_enabled_jobs')
    mocks.attach_mock(mocker.patch.object(tui, '_update_jobs_container'), '_update_jobs_container')
    mocks.attach_mock(mocker.patch.object(tui._app, 'invalidate'), 'invalidate')

    tui._refresh_jobs()

    assert mocks.mock_calls == [
        call._start_enabled_jobs(),
        call._update_jobs_container(),
        call.invalidate(),
    ]


def test__start_enabled_jobs(tui, mocker):
    mocks = Mock()
    jobinfos = {
        'a': JobInfo(job=Job(name='a', was_started=False, autostart=False, start=mocks.start_a)),
        'b': JobInfo(job=Job(name='b', was_started=False, autostart=True, start=mocks.start_b)),
        'c': JobInfo(job=Job(name='c', was_started=True, autostart=False, start=mocks.start_c)),
        'd': JobInfo(job=Job(name='d', was_started=True, autostart=True, start=mocks.start_d)),
        'e': JobInfo(job=Job(name='e', was_started=True, autostart=True, start=mocks.start_e)),
        'f': JobInfo(job=Job(name='f', was_started=True, autostart=False, start=mocks.start_f)),
        'g': JobInfo(job=Job(name='g', was_started=False, autostart=True, start=mocks.start_g)),
        'h': JobInfo(job=Job(name='h', was_started=False, autostart=False, start=mocks.start_h)),
    }

    mocker.patch.object(type(tui), '_enabled_jobs', tuple(jobinfos.values()))

    tui._start_enabled_jobs()

    assert mocks.mock_calls == [
        call.start_b(),
        call.start_g(),
    ]


def test__enabled_jobs(tui, mocker):
    jobinfos = {
        'a': JobInfo(job=Job(name='a', is_enabled=False)),
        'b': JobInfo(job=Job(name='b', is_enabled=False)),
        'c': JobInfo(job=Job(name='c', is_enabled=True)),
        'd': JobInfo(job=Job(name='d', is_enabled=True)),
        'e': JobInfo(job=Job(name='e', is_enabled=False)),
        'f': JobInfo(job=Job(name='f', is_enabled=True)),
    }

    mocker.patch.object(tui, '_jobs', jobinfos)

    assert tui._enabled_jobs == (
        jobinfos['c'],
        jobinfos['d'],
        jobinfos['f'],
    )


def test__all_jobs_finished(tui, mocker):
    jobinfos = {
        'a': JobInfo(job=Job(name='a', is_finished=False)),
        'b': JobInfo(job=Job(name='b', is_finished=False)),
        'c': JobInfo(job=Job(name='c', is_finished=True)),
        'd': JobInfo(job=Job(name='d', is_finished=True)),
        'e': JobInfo(job=Job(name='e', is_finished=False)),
        'f': JobInfo(job=Job(name='f', is_finished=True)),
    }
    mocker.patch.object(type(tui), '_enabled_jobs', PropertyMock(return_value=tuple(jobinfos.values())))

    assert tui._all_jobs_finished is False
    jobinfos['a'].job.is_finished = True
    assert tui._all_jobs_finished is False
    jobinfos['b'].job.is_finished = True
    assert tui._all_jobs_finished is False
    jobinfos['c'].job.is_finished = True
    assert tui._all_jobs_finished is False
    jobinfos['e'].job.is_finished = True
    assert tui._all_jobs_finished is True


@pytest.mark.parametrize(
    argnames='exception_from_jobs, exit_codes, exp_result',
    argvalues=(
        (None, [0, 0, 0], 0),
        (None, [0, 3, 4], 3),
        (RuntimeError('very bad'), [0, 0, 0], RuntimeError('very bad')),
    ),
    ids=lambda v: repr(v),
)
def test_run(exception_from_jobs, exit_codes, exp_result, tui, mocker):
    mocks = Mock(
        wait_a=AsyncMock(),
        wait_b=AsyncMock(),
        wait_c=AsyncMock(),
        wait_d=AsyncMock(),
        wait_e=AsyncMock(),
        wait_f=AsyncMock(),
        wait_g=AsyncMock(),
        wait_h=AsyncMock(),
    )
    mocks.attach_mock(mocker.patch.object(tui, 'add_jobs'), 'add_jobs')
    mocks.attach_mock(mocker.patch.object(tui, '_run'), '_run')
    mocks.attach_mock(
        mocker.patch.object(tui, '_get_exception', return_value=exception_from_jobs),
        '_get_exception',
    )
    jobinfos = {
        'a': JobInfo(job=Job(name='a', exit_code=exit_codes[0])),
        'b': JobInfo(job=Job(name='b', exit_code=exit_codes[1])),
        'c': JobInfo(job=Job(name='c', exit_code=exit_codes[2])),
    }
    mocker.patch.object(type(tui), '_enabled_jobs', tuple(jobinfos.values()))
    jobs = tuple(jobinfo.job for jobinfo in jobinfos.values())

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tui.run(jobs)
    else:
        return_value = tui.run(jobs)
        assert return_value == exp_result

    assert mocks.mock_calls == [
        call.add_jobs(*jobs),
        call._run(),
        call._get_exception(),
    ]


@pytest.mark.parametrize(
    argnames='run_async_exception, exp_jobs_terminated, exp_exception',
    argvalues=(
        (None, False, None),
        (asyncio.CancelledError(), True, None),
        (RuntimeError('oh no'), True, RuntimeError('oh no')),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__run(run_async_exception, exp_jobs_terminated, exp_exception, tui, mocker):
    mocks = Mock(
        wait_a=AsyncMock(),
        wait_b=AsyncMock(),
        wait_c=AsyncMock(),
        wait_d=AsyncMock(),
        wait_e=AsyncMock(),
        wait_f=AsyncMock(),
        wait_g=AsyncMock(),
        wait_h=AsyncMock(),
    )
    mocks.attach_mock(mocker.patch.object(tui, '_start_enabled_jobs'), '_start_enabled_jobs')
    mocks.attach_mock(mocker.patch.object(tui, '_terminate_jobs'), '_terminate_jobs')
    mocks.attach_mock(mocker.patch.object(tui._app, 'run_async'), 'run_async')
    mocks.run_async.side_effect = run_async_exception
    jobinfos = {
        'a': JobInfo(job=Job(name='a', was_started=False, is_finished=False, wait=mocks.wait_a)),
        'b': JobInfo(job=Job(name='b', was_started=False, is_finished=True, wait=mocks.wait_b)),
        'c': JobInfo(job=Job(name='c', was_started=True, is_finished=False, wait=mocks.wait_c)),
        'd': JobInfo(job=Job(name='d', was_started=True, is_finished=True, wait=mocks.wait_d)),
        'e': JobInfo(job=Job(name='e', was_started=True, is_finished=True, wait=mocks.wait_e)),
        'f': JobInfo(job=Job(name='f', was_started=True, is_finished=False, wait=mocks.wait_f)),
        'g': JobInfo(job=Job(name='g', was_started=False, is_finished=True, wait=mocks.wait_g)),
        'h': JobInfo(job=Job(name='h', was_started=False, is_finished=False, wait=mocks.wait_h)),
    }
    mocker.patch.object(type(tui), '_enabled_jobs', tuple(jobinfos.values()))

    exp_mock_calls = [
        call._start_enabled_jobs(),
        call.run_async(set_exception_handler=False),
    ]

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await tui._run()
    else:
        await tui._run()

        if exp_jobs_terminated:
            exp_mock_calls.append(call._terminate_jobs())
        exp_mock_calls.append(call.wait_c())
        exp_mock_calls.append(call.wait_f())

    assert mocks.mock_calls == exp_mock_calls


def test__exit_when_run_task_is_already_cancelled(tui, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tui._loop, 'call_later'), 'call_later')
    tui._run_task = mocks._run_task
    tui._run_task_cancelled = True

    tui._exit()

    assert mocks.mock_calls == []

def test__exit_when_run_task_was_created(tui, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tui._loop, 'call_later'), 'call_later')
    tui._run_task = mocks._run_task
    tui._run_task_cancelled = False

    tui._exit()

    assert mocks.mock_calls == [call._run_task.cancel()]

def test__exit_when_run_task_was_not_created(tui, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tui._loop, 'call_later'), 'call_later')
    tui._run_task = None
    tui._run_task_cancelled = False

    tui._exit()

    assert mocks.mock_calls == [call.call_later(0, tui._exit)]


def test__terminate_jobs(tui, mocker):
    mocks = Mock()
    tui._jobs = {
        'a': JobInfo(job=Job(name='a', is_finished=False, terminate=mocks.terminate_a)),
        'b': JobInfo(job=Job(name='b', is_finished=False, terminate=mocks.terminate_b)),
        'c': JobInfo(job=Job(name='c', is_finished=True, terminate=mocks.terminate_c)),
        'd': JobInfo(job=Job(name='d', is_finished=True, terminate=mocks.terminate_d)),
        'e': JobInfo(job=Job(name='e', is_finished=False, terminate=mocks.terminate_e)),
        'f': JobInfo(job=Job(name='f', is_finished=False, terminate=mocks.terminate_f)),
        'g': JobInfo(job=Job(name='g', is_finished=True, terminate=mocks.terminate_g)),
        'h': JobInfo(job=Job(name='h', is_finished=True, terminate=mocks.terminate_h)),
    }

    tui._terminate_jobs()

    assert mocks.mock_calls == [
        call.terminate_a(),
        call.terminate_b(),
        call.terminate_e(),
        call.terminate_f(),
    ]


@pytest.mark.parametrize(
    argnames='uncaught_exception, jobs, exp_exception',
    argvalues=(
        (
            RuntimeError('Application went haywire'),
            {
                'a': JobInfo(job=Job(name='a', raised=None)),
                'c': JobInfo(job=Job(name='b', raised=ValueError('wrong value, mate'))),
                'f': JobInfo(job=Job(name='c', raised=None)),
            },
            RuntimeError('Application went haywire'),
        ),
        (
            None,
            {
                'a': JobInfo(job=Job(name='a', raised=None)),
                'b': JobInfo(job=Job(name='b', raised=None)),
                'c': JobInfo(job=Job(name='c', raised=ValueError('wrong value, mate'))),
                'd': JobInfo(job=Job(name='d', raised=None)),
                'e': JobInfo(job=Job(name='e', raised=TypeError('wrong type, dude'))),
                'f': JobInfo(job=Job(name='f', raised=None)),
            },
            ValueError('wrong value, mate'),
        ),
        (
            None,
            {
                'a': JobInfo(job=Job(name='a', raised=None)),
                'b': JobInfo(job=Job(name='b', raised=None)),
                'c': JobInfo(job=Job(name='c', raised=None)),
            },
            None,
        ),
    ),
    ids=lambda v: repr(v),
)
def test__get_exception(uncaught_exception, jobs, exp_exception, tui):
    tui._jobs = jobs
    tui._exception = uncaught_exception

    exception = tui._get_exception()
    if exp_exception:
        assert type(exception) is type(exp_exception)
        assert str(exception) is str(exp_exception)
    else:
        assert exception is None



import asyncio
from unittest.mock import AsyncMock, call

import pytest

from upsies import errors
from upsies.jobs.custom import CustomJob


@pytest.fixture
def make_CustomJob(tmp_path):
    def make_CustomJob(**kwargs):
        return CustomJob(home_directory=tmp_path, cache_directory=tmp_path, **kwargs)
    return make_CustomJob


def test_name(make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    assert job.name == 'foo'


def test_label(make_CustomJob):
    job = make_CustomJob(name='foo', label='Foo', worker=AsyncMock())
    assert job.label == 'Foo'


def test_initialize(make_CustomJob):
    worker = AsyncMock()
    job = make_CustomJob(
        name='foo',
        label='Foo',
        worker=worker,
        catch=(TypeError, ValueError),
    )
    assert job._worker is worker
    assert job._expected_exceptions == (TypeError, ValueError)


@pytest.mark.asyncio
async def test_worker_raises_unexpected_exception(mocker, make_CustomJob):
    worker = AsyncMock(side_effect=RuntimeError('Oh no!'))
    job = make_CustomJob(name='foo', label='Foo', worker=worker, catch=(ValueError,))
    mocker.patch.object(job, 'error')

    job.start()
    with pytest.raises(RuntimeError, match=r'^Oh no!$'):
        await job.wait()

    assert worker.call_args_list == [call(job)]
    assert job.error.call_args_list == []
    assert job.output == ()
    assert job.is_finished is True
    assert job.exit_code == 1


@pytest.mark.asyncio
async def test_worker_raises_expected_exception(mocker, make_CustomJob):
    worker = AsyncMock(side_effect=errors.RequestError('Oh oh.'))
    job = make_CustomJob(name='foo', label='Foo', worker=worker, catch=(errors.RequestError,))
    mocker.patch.object(job, 'error')

    job.start()
    await job.wait()

    assert worker.call_args_list == [call(job)]
    assert job.error.call_args_list == [call(errors.RequestError('Oh oh.'))]
    assert job.output == ()
    assert job.is_finished is True
    assert job.exit_code == 1


@pytest.mark.asyncio
async def test_worker_returns_None(make_CustomJob, mocker):
    worker = AsyncMock(return_value=None)
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    mocker.patch.object(job, 'error')

    job.start()
    await job.wait()

    assert worker.call_args_list == [call(job)]
    assert job.error.call_args_list == []
    assert job.output == ()
    assert job.is_finished is True
    assert job.exit_code == 1


@pytest.mark.asyncio
async def test_worker_returns_iterable(make_CustomJob, mocker):
    worker = AsyncMock(return_value=('foo', 'bar', 'baz'))
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    mocker.patch.object(job, 'error')

    job.start()
    await job.wait()

    assert worker.call_args_list == [call(job)]
    assert job.error.call_args_list == []
    assert job.output == ('foo', 'bar', 'baz')
    assert job.is_finished is True
    assert job.exit_code == 0


@pytest.mark.parametrize('return_value', ('look at my work', ''))
@pytest.mark.asyncio
async def test_worker_returns_noniterable(return_value, make_CustomJob, mocker):
    worker = AsyncMock(return_value=return_value)
    job = make_CustomJob(name='foo', label='Foo', worker=worker)
    mocker.patch.object(job, 'error')

    job.start()
    await job.wait()

    assert worker.call_args_list == [call(job)]
    assert job.error.call_args_list == []
    assert job.output == (return_value,)
    assert job.is_finished is True
    assert job.exit_code == 0


@pytest.mark.asyncio
async def test_terminate_cancels_worker(make_CustomJob, mocker):
    async def delay(job):
        await asyncio.sleep(30)
        return 'sorry for being late'

    job = make_CustomJob(name='foo', label='Foo', worker=delay)
    mocker.patch.object(job, 'error')

    job.start()
    asyncio.get_running_loop().call_later(0.1, job.terminate)
    await job.wait()

    assert job.error.call_args_list == []
    assert job.output == ()
    assert job.exit_code == 1
    assert job.is_finished is True



import re
from unittest.mock import Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.set import SetJob


def test_name():
    assert SetJob.name == 'set'


def test_label():
    assert SetJob.label == 'Set'


def test_hidden():
    assert SetJob.hidden is True


def test_cache_id():
    assert SetJob.cache_id is None


@pytest.mark.parametrize(
    argnames='kwargs, exp_exception, exp_attributes',
    argvalues=(
        pytest.param(
            {
                'config': 'mock config',
            },
            None,
            {
                '_config': 'mock config',
                '_option': None,
                '_value': '',
                '_reset': None,
                '_dump': (),
            },
            id='Display complete configuration',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'option': 'mock option',
            },
            None,
            {
                '_config': 'mock config',
                '_option': 'mock option',
                '_value': '',
                '_reset': None,
                '_dump': (),
            },
            id='Display single option',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'option': 'mock option',
                'value': 'my value',
            },
            None,
            {
                '_config': 'mock config',
                '_option': 'mock option',
                '_value': 'my value',
                '_reset': None,
                '_dump': (),
            },
            id='Change option',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'option': 'mock option',
                'reset': True,
            },
            None,
            {
                '_config': 'mock config',
                '_option': 'mock option',
                '_value': '',
                '_reset': True,
                '_dump': (),
            },
            id='Reset complete configuration',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'reset': True,
            },
            None,
            {
                '_config': 'mock config',
                '_option': None,
                '_value': '',
                '_reset': True,
                '_dump': (),
            },
            id='Reset single option',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'dump': ('this', 'and', 'that'),
            },
            None,
            {
                '_config': 'mock config',
                '_option': None,
                '_value': '',
                '_reset': None,
                '_dump': ('this', 'and', 'that'),
            },
            id='Dump configuration',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'option': 'mock option',
                'dump': ('this', 'and', 'that'),
            },
            RuntimeError('Arguments "option" and "dump" are mutually exclusive.'),
            {},
            id='Invalid arguments: option and dump',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'value': 'my value',
                'reset': True,
            },
            RuntimeError('Arguments "value" and "reset" are mutually exclusive.'),
            {},
            id='Invalid arguments: value and reset',
        ),

        pytest.param(
            {
                'config': 'mock config',
                'value': 'my value',
                'dump': ('this', 'and', 'that'),
            },
            RuntimeError('Arguments "value" and "dump" are mutually exclusive.'),
            {},
            id='Invalid arguments: value and dump',
        ),
    ),
    ids=lambda v: repr(v),
)
def test_initialize(kwargs, exp_exception, exp_attributes, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(SetJob, 'run'), 'run')

    if isinstance(exp_exception, Exception):
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            SetJob(**kwargs)
    else:
        job = SetJob(**kwargs)
        assert {
            name: getattr(job, name)
            for name in exp_attributes
        } == exp_attributes


@pytest.mark.parametrize(
    argnames='attributes, methods, exp_mock_calls',
    argvalues=(
        pytest.param(
            {'_reset': True, '_value': 'my value', '_dump': ('this', 'and', 'that')},
            {'_reset_mode': Mock()},
            [
                call._reset_mode(),
            ],
            id='Reset mode succeeds',
        ),
        pytest.param(
            {'_reset': True, '_value': 'my value', '_dump': ('this', 'and', 'that')},
            {'_reset_mode': Mock(side_effect=errors.ConfigError('reset error'))},
            [
                call._reset_mode(),
                call.error(errors.ConfigError('reset error')),
            ],
            id='Reset mode fails',
        ),

        pytest.param(
            {'_reset': False, '_value': 'my value', '_dump': ('this', 'and', 'that')},
            {'_set_mode': Mock()},
            [
                call._set_mode(),
            ],
            id='Set mode succeeds',
        ),
        pytest.param(
            {'_reset': False, '_value': 'my value', '_dump': ('this', 'and', 'that')},
            {'_set_mode': Mock(side_effect=errors.ConfigError('set error'))},
            [
                call._set_mode(),
                call.error(errors.ConfigError('set error')),
            ],
            id='Set mode fails',
        ),

        pytest.param(
            {'_reset': False, '_value': '', '_dump': ('this', 'and', 'that')},
            {'_dump_mode': Mock()},
            [
                call._dump_mode(),
            ],
            id='Dump mode succeeds',
        ),
        pytest.param(
            {'_reset': False, '_value': '', '_dump': ('this', 'and', 'that')},
            {'_dump_mode': Mock(side_effect=errors.ConfigError('dump error'))},
            [
                call._dump_mode(),
                call.error(errors.ConfigError('dump error')),
            ],
            id='Dump mode fails',
        ),

        pytest.param(
            {'_reset': False, '_value': '', '_dump': ()},
            {'_display_mode': Mock()},
            [
                call._display_mode(),
            ],
            id='Display mode succeeds',
        ),
        pytest.param(
            {'_reset': False, '_value': '', '_dump': ()},
            {'_display_mode': Mock(side_effect=errors.ConfigError('display error'))},
            [
                call._display_mode(),
                call.error(errors.ConfigError('display error')),
            ],
            id='Display mode fails',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_run(attributes, methods, exp_mock_calls, mocker):
    job = SetJob(config='mock config')

    for name, value in attributes.items():
        mocker.patch.object(type(job), name, PropertyMock(return_value=value), create=True)

    mocks = Mock()
    for name, mock in methods.items():
        mocks.attach_mock(mocker.patch.object(job, name, mock), name)
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    await job.run()

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='option, paths, exp_mock_calls',
    argvalues=(
        pytest.param(
            'foo.b.that',
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call.reset('foo.b.that'),
                call._write('foo.b.that'),
            ],
            id='Reset single option',
        ),
        pytest.param(
            None,
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call.reset('foo.a.this'), call._write('foo.a.this'),
                call.reset('foo.b.that'), call._write('foo.b.that'),
                call.reset('bar.c.baz'), call._write('bar.c.baz'),
            ],
            id='Reset all options',
        ),
    ),
    ids=lambda v: repr(v),
)
def test__reset_mode(option, paths, exp_mock_calls, mocker):
    job = SetJob(config='mock config')

    mocker.patch.object(job, '_option', option)
    mocker.patch.object(job, '_config', Mock(paths=paths))
    mocker.patch.object(job, '_write', Mock())

    mocks = Mock()
    mocks.attach_mock(job._config.reset, 'reset')
    mocks.attach_mock(job._write, '_write')

    job._reset_mode()

    assert mocks.mock_calls == exp_mock_calls


def test__set_mode(mocker):
    job = SetJob(config='mock config')

    mocker.patch.object(job, '_config', {})
    mocker.patch.object(job, '_option', 'my option')
    mocker.patch.object(job, '_value', 'my value')
    mocker.patch.object(job, '_write', Mock())

    job._set_mode()
    assert job._config['my option'] == 'my value'
    assert job._write.call_args_list == [call('my option')]


@pytest.mark.parametrize(
    argnames='option, paths, exp_mock_calls',
    argvalues=(
        pytest.param(
            'foo.b.that',
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call._display_option('foo.b.that'),
            ],
            id='Display single option',
        ),
        pytest.param(
            '',
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call._display_option('foo.a.this'),
                call._display_option('foo.b.that'),
                call._display_option('bar.c.baz'),
            ],
            id='Display all options',
        ),
    ),
    ids=lambda v: repr(v),
)
def test__display_mode(option, paths, exp_mock_calls, mocker):
    job = SetJob(config='mock config')

    mocker.patch.object(job, '_option', option)
    mocker.patch.object(job, '_config', Mock(paths=paths))

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(job, '_display_option', Mock()),
        '_display_option',
    )

    job._display_mode()

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='option, paths, exp_mock_calls',
    argvalues=(
        pytest.param(
            'foo.b.that',
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call._display_option('foo.b.that'),
            ],
            id='Display single option',
        ),
        pytest.param(
            '',
            ('foo.a.this', 'foo.b.that', 'bar.c.baz'),
            [
                call._display_option('foo.a.this'),
                call._display_option('foo.b.that'),
                call._display_option('bar.c.baz'),
            ],
            id='Display all options',
        ),
    ),
    ids=lambda v: repr(v),
)
def test__dump_mode(option, paths, exp_mock_calls, mocker, tmp_path):
    job = SetJob(config='mock config')

    files = {
        'foo': tmp_path / 'foo.ini',
        'bar': tmp_path / 'bar.ini',
        'baz': tmp_path / 'baz.ini',
    }

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocker.patch.object(job, '_dump', ('foo', 'baz'))

    def write(section, include_defaults):
        files[section].write_text(f'[mock section {section}, include_defaults={include_defaults}')

    mocker.patch.object(job, '_config', Mock(
        files=files,
        write=Mock(side_effect=write),
    ))
    mocks.attach_mock(job._config.write, '_config_write')

    job._dump_mode()

    assert mocks.mock_calls == [
        call._config_write('foo', include_defaults=True),
        call.send('#' * 64),
        call.send('### ' + (str(files['foo']) + ' ').ljust(60, '#')),
        call.send('#' * 64),
        call.send('[mock section foo, include_defaults=True'),

        call._config_write('baz', include_defaults=True),
        call.send('#' * 64),
        call.send('### ' + (str(files['baz']) + ' ').ljust(60, '#')),
        call.send('#' * 64),
        call.send('[mock section baz, include_defaults=True'),
    ]

    assert files['foo'].read_text() == '[mock section foo, include_defaults=True'
    assert not files['bar'].exists()
    assert files['baz'].read_text() == '[mock section baz, include_defaults=True'


def test__write(mocker):
    job = SetJob(config='mock config')

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_config', Mock(write=Mock())), '_config')
    mocks.attach_mock(mocker.patch.object(job, '_display_option', Mock()), '_display_option')

    job._write('my option')
    assert mocks.mock_calls == [
        call._config.write('my option'),
        call._display_option('my option'),
    ]


@pytest.mark.parametrize(
    argnames='option, value, exp_mock_calls',
    argvalues=(
        pytest.param(
            'my option',
            ('foo', 'bar', 'baz'),
            [
                call.send('my option =\n  ' + 'foo\n  bar\n  baz')
            ],
            id='Non-empty sequence',
        ),
        pytest.param(
            'my option',
            (),
            [
                call.send('my option =')
            ],
            id='Empty sequence',
        ),
        pytest.param(
            'my option',
            'hello',
            [
                call.send('my option = hello')
            ],
            id='Non-sequence',
        ),
    ),
    ids=lambda v: repr(v),
)
def test__display_option(option, value, exp_mock_calls, mocker):
    job = SetJob(config='mock config')

    mocker.patch.object(job, '_config', {option: value})

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')

    job._display_option(option)
    assert mocks.mock_calls == exp_mock_calls



import asyncio
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import __project_name__, errors
from upsies.trackers import base


def make_MockTracker(**kwargs):
    class MockTracker(base.TrackerBase):
        name = 'asdf'
        label = 'AsdF'
        TrackerJobs = PropertyMock()
        TrackerConfig = PropertyMock()
        _login = AsyncMock()
        _logout = AsyncMock()
        get_announce_url = AsyncMock()
        upload = AsyncMock()

    return MockTracker(**kwargs)


def test_generate_setup_howto(mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.trackers.base.tracker.Howto'), 'Howto')
    mocks.attach_mock(mocker.patch('upsies.utils.string.evaluate_fstring'), 'evaluate_fstring')
    tracker = make_MockTracker()
    return_value = type(tracker).generate_setup_howto()
    assert return_value is mocks.evaluate_fstring.return_value
    assert mocks.mock_calls == [
        call.Howto(tracker_cls=type(tracker)),
        call.evaluate_fstring(
            type(tracker).setup_howto_template,
            howto=mocks.Howto.return_value,
            tracker=type(tracker),
            executable=__project_name__,
        ),
    ]


def test_options():
    options = {'username': 'foo', 'password': 'bar'}
    tracker = make_MockTracker(options=options)
    assert tracker.options is options


def test_tasks():
    tracker = make_MockTracker()
    assert tracker._tasks == []
    assert tracker._tasks is tracker._tasks


@pytest.mark.parametrize('callback', (None, Mock()), ids=('without callback', 'with callback'))
@pytest.mark.asyncio
async def test_attach_task(callback, mocker):
    tracker = make_MockTracker()

    work = AsyncMock()
    if callback:
        task = tracker.attach_task(work(), callback=callback)
    else:
        task = tracker.attach_task(work())

    assert task in tracker._tasks

    await task

    assert task not in tracker._tasks
    if callback:
        assert callback.call_args_list == [call(task)]


@pytest.mark.asyncio
async def test_await_tasks(mocker):
    tracker = make_MockTracker()
    tasks = (
        asyncio.ensure_future(AsyncMock()()),
        asyncio.ensure_future(AsyncMock()()),
        asyncio.ensure_future(AsyncMock()()),
    )
    tracker._tasks.extend(tasks)

    for task in tasks:
        assert not task.done()

    await tracker.await_tasks()

    for task in tasks:
        assert task.done()


@pytest.mark.asyncio
async def test_login_when_already_logged_in(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_login')
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Login multiple times concurrently
    await asyncio.gather(
        tracker.login(),
        tracker.login(),
        tracker.login(),
    )

    assert tracker.is_logged_in
    assert tracker._login.call_args_list == []

@pytest.mark.asyncio
async def test_login_succeeds(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_login')

    assert not tracker.is_logged_in

    # Login multiple times concurrently
    await asyncio.gather(
        tracker.login(),
        tracker.login(),
        tracker.login(),
    )

    assert tracker.is_logged_in
    assert tracker._login.call_args_list == [call()]

@pytest.mark.asyncio
async def test_login_fails(mocker):
    tracker = make_MockTracker()
    exception = errors.RequestError('nope')
    mocker.patch.object(tracker, '_login', side_effect=exception)

    assert not tracker.is_logged_in

    # Login multiple times concurrently
    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await asyncio.gather(
            tracker.login(),
            tracker.login(),
            tracker.login(),
        )

    assert not tracker.is_logged_in
    assert tracker._login.call_args_list == [call(), call(), call()]

@pytest.mark.asyncio
async def test_logout_when_already_logged_out(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_logout')
    mocker.patch.object(tracker, '_is_logged_in', False, create=True)

    assert not tracker.is_logged_in

    # Logout multiple times concurrently
    await asyncio.gather(
        tracker.logout(),
        tracker.logout(),
        tracker.logout(),
    )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == []

async def test_logout_succeeds(mocker):
    tracker = make_MockTracker()
    mocker.patch.object(tracker, '_logout')
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Logout multiple times concurrently
    await asyncio.gather(
        tracker.logout(),
        tracker.logout(),
        tracker.logout(),
    )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == [call()]

@pytest.mark.asyncio
async def test_logout_fails(mocker):
    tracker = make_MockTracker()
    exception = errors.RequestError('nope')
    mocker.patch.object(tracker, '_logout', side_effect=exception)
    mocker.patch.object(tracker, '_is_logged_in', True, create=True)

    assert tracker.is_logged_in

    # Logout multiple times concurrently
    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await asyncio.gather(
            tracker.logout(),
            tracker.logout(),
            tracker.logout(),
        )

    assert not tracker.is_logged_in
    assert tracker._logout.call_args_list == [call()]


def test_calculate_piece_size():
    tracker = make_MockTracker()
    assert type(tracker).calculate_piece_size is None


def test_calculate_piece_size_min_max():
    tracker = make_MockTracker()
    assert type(tracker).calculate_piece_size_min_max is None


def test_signals():
    tracker = make_MockTracker()
    cb = Mock()
    tracker.signal.register('warning', cb.warn)
    tracker.signal.register('error', cb.error)
    tracker.signal.register('exception', cb.exception)
    tracker.warn('foo')
    assert cb.mock_calls == [call.warn('foo')]
    tracker.error('bar')
    assert cb.mock_calls == [call.warn('foo'), call.error('bar')]
    tracker.exception('baz')
    assert cb.mock_calls == [call.warn('foo'), call.error('bar'), call.exception('baz')]



from itertools import zip_longest

import pytest

from upsies.utils.types import ReleaseType
from upsies.utils.webdbs import Query, SearchResult, imdb


@pytest.fixture
def api():
    return imdb.ImdbApi()


ids = {
    'pygmalion': 'tt0190673',                       # Actor name contains "&apos;"
    '1899': 'tt9319668',                            # Series with digit-only title
    '1984': 'tt0048918',                            # Movie with digit-only title
    '3 days to kill': 'tt2172934',                  # Movie title starts with number
    'adu': 'tt9616700',                             # Movie title contains special character (Adú)
    'the nest': 'tt0280990',                        # Non-English AKA contains special character (guêpes)
    'february': 'tt3286052',                        # English AKA
    'ao': 'tt1526578',                              # No English AKA available
    'wind in the willows': 'tt0192802',             # TV movie
    'benders big score': 'tt0471711',               # Video
    'elephant': 'tt0097270',                        # TV movie
    'kung fury': 'tt3472226',                       # Short
    'the forest': 'tt6560040',                      # TV miniseries
    'the bridge': 'tt1733785',                      # Series
    'the bridge s02e03': 'tt3146200',               # Episode
    'aftermath': 'tt0896516',                       # Video, no cast list
    'wingwall auditions': 'tt0253248',              # No description
    'the believer': 'tt0247199',                    # Single-word genre
    'watchmen': 'tt0409459',                        # Runtimes: Director's Cut and Ultimate Cut
    'thale 2': 'tt2917940',                         # Unreleased movie

    # Title AKA regression tests
    'butterfly': 'tt0188030',
    'dead and buried': 'tt0082242',
    'hard boiled': 'tt0104684',
    'joy ride 3': 'tt3138376',
    'pusher 2': 'tt0396184',
    'sin nombre': 'tt1127715',
    'terror in the woods': 'tt7534328',
    'the human centipede 2': 'tt1530509',
}


@pytest.mark.parametrize(
    argnames='query, exp_query',
    argvalues=(
        (Query('Foo and Bar'), Query('foo bar')),
        (Query('Foo & Bar'), Query('foo & bar')),
        (Query('Cant See You'), Query("can't See You")),
        (Query('Wont See You'), Query("Won't See You")),
        (Query('Dont See You'), Query("Don't See You")),
        (Query('Want to See You'), Query("Want to See You")),
    ),
    ids=lambda value: str(value),
)
def test_sanitize_query(query, exp_query, api, store_response):
    return_value = api.sanitize_query(query)
    assert return_value == exp_query


@pytest.mark.parametrize(
    argnames='text, exp_id',
    argvalues=(
        ('Some Title', None),
        ('1984', None),
        ('tt0123arf', None),
        ('arftt0123', None),
        ('tt0123', 'tt0123'),
        ('https://www.imdb.com/title/tt0123/', 'tt0123'),
        ('https://www.imdb.com/title/tt0123', 'tt0123'),
        ('junk tt0123 more junk', 'tt0123'),
        ('junk https://www.imdb.com/title/tt0123/stuff more junk', 'tt0123'),
    ),
    ids=lambda value: str(value),
)
def test_get_id_from_text(text, exp_id, api, store_response):
    return_value = api.get_id_from_text(text)
    assert return_value == exp_id


@pytest.mark.parametrize(
    argnames='query, title, title_english, title_original, type, year, cast, countries, directors, genres, summary',
    argvalues=(
        (
            Query(id=ids['february']),
            "The Blackcoat's Daughter",
            "The Blackcoat's Daughter",
            'February',
            ReleaseType.movie,
            '2015',
            ('Emma Roberts', 'Kiernan Shipka', 'Lucy Boynton'),
            ('Canada', 'United States'),
            ('Oz Perkins',),
            ('horror', 'mystery', 'thriller'),
            ('Two girls must battle a mysterious evil force when they get left behind '
             'at their boarding school over winter break.'),
        ),
        (
            Query(id=ids['the bridge']),
            'The Bridge',
            'The Bridge',
            'Bron/Broen',
            ReleaseType.series,
            '2011',
            ('Sofia Helin', 'Rafael Pettersson', 'Sarah Boberg'),
            ('Sweden', 'Denmark', 'Germany', 'Norway'),
            (),
            ('crime', 'mystery', 'thriller'),
            ('When a body is found on the bridge between Denmark and Sweden, '
             'right on the border, Danish inspector Martin Rohde and Swedish '
             'Saga Norén have to share jurisdiction and work together to find the killer.'),
        ),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_handles_id_in_query(query, title, title_english, title_original, type, year,
                                          cast, countries, directors, genres, summary,
                                          api, store_response):
    results = await api.search(query)
    assert len(results) == 1
    assert results[0].id == query.id
    assert results[0].title == title
    assert results[0].type == type
    assert results[0].url == f'{imdb.ImdbApi._url_base}/title/{query.id}'
    assert results[0].year == year
    assert (await results[0].cast())[:3] == cast
    assert (await results[0].countries()) == countries
    assert (await results[0].directors()) == directors
    assert (await results[0].genres()) == genres
    assert (await results[0].summary()) == summary
    assert (await results[0].title_english()) == title_english
    assert (await results[0].title_original()) == title_original


@pytest.mark.asyncio
async def test_search_returns_empty_list_if_title_is_empty(api, store_response):
    assert await api.search(Query('', year='2009')) == []

@pytest.mark.asyncio
async def test_search_returns_list_of_SearchResults(api, store_response):
    results = await api.search(Query('blues brothers', type=ReleaseType.movie))
    for result in results:
        assert isinstance(result, SearchResult)


@pytest.mark.parametrize(
    argnames=('query', 'exp_ids'),
    argvalues=(
        (Query('blues brothers', year=1980), ('tt0080455',)),  # The Blues Brothers
        (Query('blues brothers', year='1998'), ('tt0118747',)),  # Blues Brothers 2000 (1998)
        (Query('blues brothers', year=1950), None),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_year(query, exp_ids, api, store_response):
    results = await api.search(query)
    ids = [r.id for r in results]
    if exp_ids:
        for exp_id in exp_ids:
            assert exp_id in ids
    else:
        assert not results


@pytest.mark.parametrize(
    argnames=('query', 'exp_ids'),
    argvalues=(
        (Query('blues brothers', type=ReleaseType.movie), ('tt0080455',)),   # 1980 movie
        (Query('blues brothers', type=ReleaseType.series), ('tt0472243',)),  # Animated series
        (Query('balada triste de trompeta', type=ReleaseType.series), None),
        (Query('time traveling bong', type=ReleaseType.movie), None),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_type(query, exp_ids, api, store_response):
    results = await api.search(query)
    ids = [r.id for r in results]
    if exp_ids:
        for exp_id in exp_ids:
            assert exp_id in ids
    else:
        assert not results


@pytest.mark.parametrize(
    argnames=('query', 'exp_cast'),
    argvalues=(
        (Query(id=ids['february']), ('Emma Roberts', 'Kiernan Shipka', 'Lucy Boynton')),
        (Query(id=ids['the bridge']), ('Sofia Helin', 'Rafael Pettersson', 'Sarah Boberg')),
        (Query(id=ids['the forest']), ('Samuel Labarthe', 'Suzanne Clément', 'Alexia Barlier')),
        (Query(id=ids['aftermath']), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_cast(query, exp_cast, api, store_response):
    results = await api.search(query)
    cast = await results[0].cast()
    for member in exp_cast:
        assert member in cast


@pytest.mark.parametrize(
    argnames=('query', 'exp_countries'),
    argvalues=(
        (Query(id=ids['february']), ('Canada', 'United States')),
        (Query(id=ids['the bridge']), ('Sweden', 'Denmark', 'Germany', 'Norway')),
        (Query(id=ids['the forest']), ('France',)),
        (Query(id=ids['aftermath']), ('United Kingdom',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_country(query, exp_countries, api, store_response):
    results = await api.search(query)
    countries = await results[0].countries()
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('query', 'exp_id'),
    argvalues=(
        (Query(id=ids['february']), ids['february']),
        (Query(id=ids['the bridge']), ids['the bridge']),
        (Query(id=ids['the forest']), ids['the forest']),
        (Query(id=ids['aftermath']), ids['aftermath']),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_id(query, exp_id, api, store_response):
    results = await api.search(query)
    assert results[0].id == exp_id


@pytest.mark.parametrize(
    argnames=('query', 'exp_directors'),
    argvalues=(
        (Query(id=ids['february']), ('Oz Perkins',)),
        (Query(id=ids['the bridge']), ()),
        (Query(id=ids['the forest']), ()),
        (Query(id=ids['aftermath']), ('Nancy Diuguid',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_directors(query, exp_directors, api, store_response):
    results = await api.search(query)
    directors = await results[0].directors()
    assert directors == exp_directors


@pytest.mark.parametrize(
    argnames=('query', 'exp_genres'),
    argvalues=(
        (Query(id=ids['february']), ('horror', 'mystery', 'thriller')),
        (Query(id=ids['the bridge']), ('crime', 'mystery', 'thriller')),
        (Query(id=ids['the forest']), ('crime', 'drama')),
        (Query(id=ids['aftermath']), ('short', 'drama')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_genres(query, exp_genres, api, store_response):
    results = await api.search(query)
    genres = await results[0].genres()
    assert genres == exp_genres


@pytest.mark.parametrize(
    argnames=('query', 'exp_summary'),
    argvalues=(
        (Query(id=ids['february']), ('Two girls must battle a mysterious evil force when they '
                                     'get left behind at their boarding school over winter break.')),
        (Query(id=ids['the bridge']), ('When a body is found on the bridge between Denmark and '
                                       'Sweden, right on the border, Danish inspector Martin Rohde '
                                       'and Swedish Saga Norén have to share jurisdiction and work '
                                       'together to find the killer.')),
        (Query(id=ids['the forest']), ('Sixteen-year-old Jennifer disappears one night from her '
                                       'village in the Ardennes. Captain Gaspard Deker leads the '
                                       'investigation with local cop Virginie Musso, who knew '
                                       'the girl well. They are helped by Eve, a lonely and mysterious woman.')),
        (Query(id=ids['wingwall auditions']), ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_summary(query, exp_summary, api, store_response):
    results = await api.search(query)
    summary = await results[0].summary()
    assert summary == exp_summary


@pytest.mark.parametrize(
    argnames='query, exp_title, exp_title_english, exp_title_original',
    argvalues=(
        (Query(id=ids['1899']), '1899', '', '1899'),
        (Query(id=ids['1984']), '1984', '', '1984'),
        (Query(id=ids['adu']), 'Adú', '', 'Adú'),
        (Query(id=ids['dead and buried']), 'Dead & Buried', '', 'Dead & Buried'),
        (Query(id=ids['february']), "The Blackcoat's Daughter", "The Blackcoat's Daughter", 'February'),
        (Query(id=ids['hard boiled']), 'Hard Boiled', 'Hard Boiled', 'Lat sau san taam'),
        (Query(id=ids['sin nombre']), 'Sin nombre', '', 'Sin nombre'),
        (Query(id=ids['the bridge']), 'The Bridge', 'The Bridge', 'Bron/Broen'),
        (Query(id=ids['the forest']), 'The Forest', 'The Forest', 'La forêt'),
        (Query(id=ids['aftermath']), 'Aftermath', '', 'Aftermath'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title_english_original(query, exp_title, exp_title_english, exp_title_original, api, store_response):
    results = await api.search(query)
    assert results[0].title == exp_title
    assert (await results[0].title_english()) == exp_title_english
    assert (await results[0].title_original()) == exp_title_original


@pytest.mark.parametrize(
    argnames=('query', 'exp_type'),
    argvalues=(
        (Query(id=ids['february']), ReleaseType.movie),
        (Query(id=ids['the bridge']), ReleaseType.season),
        (Query(id=ids['the forest']), ReleaseType.season),
        (Query(id=ids['aftermath']), ReleaseType.movie),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_type(query, exp_type, api, store_response):
    results = await api.search(query)
    assert results[0].type == exp_type


@pytest.mark.parametrize(
    argnames=('query', 'exp_url'),
    argvalues=(
        (Query(id=ids['february']), 'https://www.imdb.com/title/tt3286052'),
        (Query(id=ids['the bridge']), 'https://www.imdb.com/title/tt1733785'),
        (Query(id=ids['the forest']), 'https://www.imdb.com/title/tt6560040'),
        (Query(id=ids['aftermath']), 'https://www.imdb.com/title/tt0896516'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_url(query, exp_url, api, store_response):
    results = await api.search(query)
    assert results[0].url == exp_url


@pytest.mark.parametrize(
    argnames=('query', 'exp_year'),
    argvalues=(
        (Query(id=ids['february']), '2015'),
        (Query(id=ids['the bridge']), '2011'),
        (Query(id=ids['the forest']), '2017'),
        (Query(id=ids['wingwall auditions']), ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_year(query, exp_year, api, store_response):
    results = await api.search(query)
    assert results[0].year == exp_year


@pytest.mark.asyncio
async def test_search_result_parser_failure(api):
    result = imdb._ImdbSearchResult(imdb_api=api)
    assert result.id == ''
    assert result.title == ''
    assert result.type == ReleaseType.movie
    assert result.url == ''
    assert result.year == ''
    assert await result.cast() == ()
    assert await result.countries() == ()
    assert await result.directors() == ()
    assert await result.genres() == ()
    assert await result.summary() == ''
    assert await result.title_english() == ''
    assert await result.title_original() == ''


@pytest.mark.parametrize(
    argnames=('id', 'exp_cast'),
    argvalues=(
        (ids['pygmalion'], (("Peter O'Toole", 'https://www.imdb.com/name/nm0000564'),
                            ('Shelagh McLeod', 'https://www.imdb.com/name/nm0572870'),
                            ('Nancy Kerr', 'https://www.imdb.com/name/nm0449770'))),
        (ids['wind in the willows'], (('Alan Bennett', 'https://www.imdb.com/name/nm0003141'),
                                      ('Michael Palin', 'https://www.imdb.com/name/nm0001589'),
                                      ('Michael Gambon', 'https://www.imdb.com/name/nm0002091'))),
        (ids['benders big score'], (('Billy West', 'https://www.imdb.com/name/nm0921942'),
                                    ('Katey Sagal', 'https://www.imdb.com/name/nm0005408'),
                                    ('John DiMaggio', 'https://www.imdb.com/name/nm0224007'))),
        (ids['elephant'], (('Gary Walker', 'https://www.imdb.com/name/nm1281884'),
                           ('Bill Hamilton', 'https://www.imdb.com/name/nm1685093'),
                           ('Michael Foyle', 'https://www.imdb.com/name/nm0289432'))),
        (ids['kung fury'], (('David Sandberg', 'https://www.imdb.com/name/nm6247887'),
                            ('Jorma Taccone', 'https://www.imdb.com/name/nm1672246'),
                            ('Steven Chew', 'https://www.imdb.com/name/nm7320022'))),
        (ids['the forest'], (('Samuel Labarthe', 'https://www.imdb.com/name/nm0479355'),
                             ('Suzanne Clément', 'https://www.imdb.com/name/nm0167501'),
                             ('Alexia Barlier', 'https://www.imdb.com/name/nm1715145'))),
        (ids['the bridge'], (('Sofia Helin', 'https://www.imdb.com/name/nm0375138'),
                             ('Rafael Pettersson', 'https://www.imdb.com/name/nm1142392'),
                             ('Sarah Boberg', 'https://www.imdb.com/name/nm0090360'))),
        (ids['the bridge s02e03'], (('Sofia Helin', 'https://www.imdb.com/name/nm0375138'),
                                    ('Kim Bodnia', 'https://www.imdb.com/name/nm0091035'),
                                    ('Dag Malmberg', 'https://www.imdb.com/name/nm0540216'))),
        (ids['aftermath'], ()),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_cast(id, exp_cast, api, store_response):
    cast = (await api.cast(id))[:3]
    if not cast:
        assert exp_cast == ()
    else:
        for person, (name, url) in zip_longest(cast, exp_cast):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_countries'),
    argvalues=(
        (ids['february'], ('Canada', 'United States')),  # Movie
        (ids['wind in the willows'], ('United Kingdom',)),  # TV Movie
        (ids['benders big score'], ('United States',)),  # Video
        (ids['elephant'], ('United Kingdom',)),  # TV movie
        (ids['kung fury'], ('Sweden',)),  # Short
        (ids['the bridge'], ('Sweden', 'Denmark', 'Germany', 'Norway')),  # Series
        (ids['the bridge s02e03'], ()),  # Episode
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_countries(id, exp_countries, api, store_response):
    countries = await api.countries(id)
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('id', 'exp_creators'),
    argvalues=(
        (ids['february'], (('Oz Perkins', 'https://www.imdb.com/name/nm0674020'),)),  # Movie
        (ids['wind in the willows'], (('Kenneth Grahame', 'https://www.imdb.com/name/nm0334370'),  # TV Movie
                                      ('Ted Walker', 'https://www.imdb.com/name/nm0908249'))),
        (ids['benders big score'], (('Matt Groening', 'https://www.imdb.com/name/nm0004981'),  # Video
                                    ('David X. Cohen', 'https://www.imdb.com/name/nm0169326'),
                                    ('Ken Keeler', 'https://www.imdb.com/name/nm0444517'))),
        (ids['elephant'], (('Bernard MacLaverty', 'https://www.imdb.com/name/nm0533735'),)),  # TV Movie
        (ids['kung fury'], (('David Sandberg', 'https://www.imdb.com/name/nm6247887'),)),  # Short
        (ids['the forest'], (('Delinda Jacobs', 'https://www.imdb.com/name/nm3064398'),)),  # Miniseries
        (ids['the bridge'], (('Måns Mårlind', 'https://www.imdb.com/name/nm0617523'),
                             ('Hans Rosenfeldt', 'https://www.imdb.com/name/nm0742577'),
                             ('Björn Stein', 'https://www.imdb.com/name/nm0825407'))),  # Series
        (ids['the bridge s02e03'], (('Hans Rosenfeldt', 'https://www.imdb.com/name/nm0742577'),  # Episode
                                    ('Måns Mårlind', 'https://www.imdb.com/name/nm0617523'),
                                    ('Björn Stein', 'https://www.imdb.com/name/nm0825407'))),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_creators(id, exp_creators, api, store_response):
    creators = await api.creators(id)
    if not creators:
        assert exp_creators == ()
    else:
        for person, (name, url) in zip_longest(creators, exp_creators, fillvalue=(None, None)):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_directors'),
    argvalues=(
        (ids['february'], (('Oz Perkins', 'https://www.imdb.com/name/nm0674020'),)),  # Movie
        (ids['wind in the willows'], (('Dave Unwin', 'https://www.imdb.com/name/nm0881386'),  # TV Movie
                                      ('Dennis Abey', 'https://www.imdb.com/name/nm0008688'))),
        (ids['benders big score'], (('Dwayne Carey-Hill', 'https://www.imdb.com/name/nm1401752'),)),  # Video
        (ids['elephant'], (('Alan Clarke', 'https://www.imdb.com/name/nm0164639'),)),  # TV movie
        (ids['kung fury'], (('David Sandberg', 'https://www.imdb.com/name/nm6247887'),)),  # Short
        (ids['the forest'], ()),  # Miniseries
        (ids['the bridge'], ()),  # Series
        (ids['the bridge s02e03'], (('Henrik Georgsson', 'https://www.imdb.com/name/nm1380082'),)),  # Episode
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_directors(id, exp_directors, api, store_response):
    directors = await api.directors(id)
    if not directors:
        assert exp_directors == ()
    else:
        for person, (name, url) in zip_longest(directors[:3], exp_directors, fillvalue=(None, None)):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_genres'),
    argvalues=(
        (ids['the believer'], ('drama',)),
        (ids['february'], ('horror', 'mystery', 'thriller')),
        (ids['wind in the willows'], ('animation', 'family')),
        (ids['benders big score'], ('animation', 'comedy', 'musical')),
        (ids['elephant'], ('crime', 'drama')),
        (ids['kung fury'], ('short', 'action', 'comedy')),
        (ids['the forest'], ('crime', 'drama')),
        (ids['the bridge'], ('crime', 'mystery', 'thriller')),
        (ids['the bridge s02e03'], ('crime', 'mystery', 'thriller')),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_genres(id, exp_genres, api, store_response):
    genres = await api.genres(id)
    assert genres == exp_genres


@pytest.mark.parametrize(
    argnames='id, exp_poster_url',
    argvalues=(
        # Movie
        (ids['february'], 'https://m.media-amazon.com/images/M/MV5BMTY2NTQ3NTEzNF5BMl5BanBnXkFtZTgwMzA0MzY1OTE@._V1_.jpg'),
        # TV movie
        (ids['wind in the willows'], 'https://m.media-amazon.com/images/M/MV5BOGQ3NWM5OTAtOThjZS00MzMyLTkwZjMtOWIyOTdlZjFiZGI1XkEyXkFqcGdeQXVyNzMwOTY2NTI@._V1_.jpg'),
        # Video
        (ids['benders big score'], 'https://m.media-amazon.com/images/M/MV5BM2U3NmRiNDItMDQwYy00Y2Q0LWJiYzItNjAzNThkZjFlM2RiXkEyXkFqcGdeQXVyNTA4NzY1MzY@._V1_.jpg'),
        # TV movie
        (ids['elephant'], 'https://m.media-amazon.com/images/M/MV5BMWE0ZTBiOWItY2ZkNS00MzI0LWE5Y2QtYjJmNjM1MGRkZGNmXkEyXkFqcGdeQXVyMzU0MTk1Nzc@._V1_.jpg'),
        # Short
        (ids['kung fury'], 'https://m.media-amazon.com/images/M/MV5BMjQwMjU2ODU5NF5BMl5BanBnXkFtZTgwNTU1NjM4NTE@._V1_.jpg'),
        # Miniseries
        (ids['the forest'], 'https://m.media-amazon.com/images/M/MV5BMzY5NThkOWItN2I1OC00MzQ2LThlYjktMGExYTk2YzM1ZGRmXkEyXkFqcGdeQXVyODEyMzI2OTE@._V1_.jpg'),
        # Series
        (ids['the bridge'], 'https://m.media-amazon.com/images/M/MV5BMjQ3MDAzNDU4NV5BMl5BanBnXkFtZTgwNjE2NDQ0NzE@._V1_.jpg'),
        # Episode
        (ids['the bridge s02e03'], 'https://m.media-amazon.com/images/M/MV5BZDc0NDY1ZDUtN2MyMy00YTg3LWJkZTktNThhYmRiYTlkOTM0L2ltYWdlXkEyXkFqcGdeQXVyNjg2MjkwNTQ@._V1_.jpg'),
        # No poster
        (ids['wingwall auditions'], ''),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_poster_url(id, exp_poster_url, api, store_response):
    poster_url = await api.poster_url(id)
    assert poster_url == exp_poster_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_rating'),
    argvalues=(
        (ids['february'], 5.9),  # Movie
        (ids['wind in the willows'], 7.4),  # TV movie
        (ids['benders big score'], 7.6),  # Video
        (ids['elephant'], 7.1),  # TV movie
        (ids['kung fury'], 8.0),  # Short
        (ids['the forest'], 7.2),  # Miniseries
        (ids['the bridge'], 8.6),  # Series
        (ids['the bridge s02e03'], 8.0),  # Episode
        (None, None),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_rating(id, exp_rating, api, store_response):
    rating = await api.rating(id)
    assert rating == exp_rating


@pytest.mark.parametrize(
    argnames=('id', 'exp_runtimes'),
    argvalues=(
        (ids['watchmen'], {'default': 162, "Director's Cut": 186, 'Ultimate Cut': 215}),  # Movie
        (ids['wind in the willows'], {'default': 73}),  # TV movie
        (ids['benders big score'], {'default': 88}),  # Video
        (ids['elephant'], {'default': 39}),  # TV movie
        (ids['kung fury'], {'default': 31}),  # Short
        (ids['the forest'], {'Entire Series': 313}),  # Miniseries
        (ids['the bridge'], {'default': 60}),  # Series
        (ids['the bridge s02e03'], {'default': 57}),  # Episode
        (ids['wingwall auditions'], {}),  # No runtime found
        (None, {}),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_runtimes(id, exp_runtimes, api, store_response):
    runtimes = await api.runtimes(id)
    assert runtimes == exp_runtimes


@pytest.mark.parametrize(
    argnames=('id', 'exp_summary'),
    argvalues=(
        (ids['february'], ('Two girls must battle a mysterious evil force when they get left '
                           'behind at their boarding school over winter break.')),  # Movie
        (ids['wind in the willows'], ('When three close friends Mole, Ratty and Badger '
                                      'find out that the infamous Mr. Toad of Toad Hall '
                                      'has been up to no good, they must find him and change '
                                      'his ways for good.')),  # TV movie
        (ids['benders big score'], ('Planet Express sees a hostile takeover and Bender falls into the hands of '
                                    'criminals where he is used to fulfill their schemes.')),  # Video
        (ids['elephant'], ('A depiction of a series of violent killings in Northern Ireland '
                           'with no clue as to exactly who is responsible.')),  # TV Short
        (ids['kung fury'], ('In 1985, Kung Fury, the toughest martial artist cop in Miami, '
                            'goes back in time to kill the worst criminal of all time '
                            '- Kung Führer, a.k.a. Adolf Hitler.')),  # Short
        (ids['the forest'], ('Sixteen-year-old Jennifer disappears one night from her village '
                             'in the Ardennes. Captain Gaspard Deker leads the investigation with '
                             'local cop Virginie Musso, who knew the girl well. They are helped '
                             'by Eve, a lonely and mysterious woman.')),  # Miniseries
        (ids['the bridge'], ('When a body is found on the bridge between Denmark and Sweden, '
                             'right on the border, Danish inspector Martin Rohde and Swedish '
                             'Saga Norén have to share jurisdiction and work together to find the killer.')),  # Series
        (ids['the bridge s02e03'], ('Saga and Martin continue their hunt for the eco-terrorists, '
                                    'who have realised the police are on their trail.')),  # Episode
        (ids['wingwall auditions'], ''),  # No summary found
        (None, ''),
    ),
    ids=lambda value: str(value)[:30] or '<empty>',
)
@pytest.mark.asyncio
async def test_summary(id, exp_summary, api, store_response):
    summary = await api.summary(id)
    if isinstance(exp_summary, str):
        assert summary == exp_summary
    else:
        assert any(summary == exp_sum for exp_sum in exp_summary)


@pytest.mark.parametrize(
    argnames=('id', 'exp_title_english', 'exp_title_original'),
    argvalues=(
        (ids['1984'], '', '1984'),
        (ids['dead and buried'], '', 'Dead & Buried'),
        (ids['hard boiled'], 'Hard Boiled', 'Lat sau san taam'),
        (ids['butterfly'], 'Butterfly', 'La lengua de las mariposas'),
        (ids['the nest'], 'The Nest', 'Nid de guêpes'),
        (ids['pusher 2'], '', 'Pusher II'),
        (ids['sin nombre'], '', 'Sin nombre'),
        (ids['ao'], '', 'Ao, le dernier Néandertal'),
        (ids['the human centipede 2'], '', 'The Human Centipede II (Full Sequence)'),
        (ids['3 days to kill'], '', '3 Days to Kill'),
        (ids['joy ride 3'], '', 'Joy Ride 3: Road Kill'),
        (ids['february'], "The Blackcoat's Daughter", 'February'),
        (ids['the forest'], 'The Forest', 'La forêt'),
        (ids['terror in the woods'], '', 'Terror in the Woods'),
        (ids['1899'], '', '1899'),
        (ids['adu'], '', 'Adú'),
        (None, '', ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_title_english_original(id, exp_title_english, exp_title_original, api, store_response):
    title_original = await api.title_original(id)
    title_english = await api.title_english(id)
    print('original:', title_original, '-- expected:', exp_title_original)
    print('english:', title_english, '-- expected:', exp_title_english)
    assert title_original == exp_title_original
    assert title_english == exp_title_english
    assert await api.title_english(id, default_to_original=False) == exp_title_english
    if exp_title_english:
        assert await api.title_english(id, default_to_original=True) == exp_title_english
    else:
        assert await api.title_english(id, default_to_original=True) == exp_title_original


@pytest.mark.parametrize(
    argnames=('id', 'exp_type'),
    argvalues=(
        (ids['february'], ReleaseType.movie),  # Movie
        (ids['wind in the willows'], ReleaseType.movie),  # TV movie
        (ids['benders big score'], ReleaseType.movie),  # Video
        (ids['elephant'], ReleaseType.movie),  # TV movie
        (ids['kung fury'], ReleaseType.movie),  # Short
        (ids['the forest'], ReleaseType.season),  # Miniseries
        (ids['the bridge'], ReleaseType.season),  # Series
        (ids['the bridge s02e03'], ReleaseType.episode),  # Episode
        (None, ReleaseType.unknown),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_type(id, exp_type, api, store_response):
    assert await api.type(id) is exp_type


@pytest.mark.asyncio
async def test_url(api):
    assert await api.url('foo') == api._url_base + '/title/foo'
    assert await api.url(None) == ''


@pytest.mark.parametrize(
    argnames=('id', 'exp_year'),
    argvalues=(
        (ids['1984'], '1956'),  # Movie with digit-only title
        (ids['1899'], '2022'),  # Series with digit-only title
        (ids['february'], '2015'),  # Movie
        (ids['wind in the willows'], '1995'),  # TV movie
        (ids['benders big score'], '2007'),  # Video
        (ids['elephant'], '1989'),  # TV movie
        (ids['kung fury'], '2015'),  # Short
        (ids['the forest'], '2017'),  # Miniseries
        (ids['the bridge'], '2011'),  # Series
        (ids['the bridge s02e03'], '2013'),  # Episode
        (ids['thale 2'], ''),  # Unreleased movie
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_year(id, exp_year, api, store_response):
    assert await api.year(id) == exp_year



import asyncio
import sys
from unittest.mock import Mock, call

import pytest

from upsies.jobs import JobBase, QueueJobBase


class FooJob(QueueJobBase):
    name = 'foo'
    label = 'Foo'

    def initialize(self, foo=None, bar=None, enqueue=()):
        self.handled_inputs = []

    async def handle_input(self, value):
        self.handled_inputs.append(value)


@pytest.fixture
def qjob(tmp_path):
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    return qjob


def test_QueueJobBase_is_JobBase_subclass(qjob):
    assert isinstance(qjob, JobBase)


@pytest.mark.skipif(sys.version_info < (3, 10), reason="fails with Python <= 3.9")
@pytest.mark.asyncio
async def test_nothing_is_enqueued(qjob):
    qjob.start()
    # Prevent wait() from blocking forever
    asyncio.get_running_loop().call_later(0.1, qjob.close)
    await qjob.wait()

    assert qjob._enqueue_args == ()
    assert qjob.handled_inputs == []


@pytest.mark.asyncio
async def test_enqueue_argument(tmp_path):
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path, enqueue=(1, 2, 3))
    qjob.start()
    await qjob.wait()

    assert qjob._enqueue_args == (1, 2, 3)
    assert qjob.handled_inputs == [1, 2, 3]


@pytest.mark.skipif(sys.version_info < (3, 10), reason="fails with Python <= 3.9")
@pytest.mark.asyncio
async def test_enqueue_method(qjob, tmp_path):
    class FeedJob(JobBase):
        name = 'feeder'
        label = 'Feeder'
        cache_directory = str(tmp_path)

        async def run(self):
            await asyncio.sleep(0.01)
            self.send('foo')
            await asyncio.sleep(0.03)
            self.send('bar')
            await asyncio.sleep(0.02)
            self.send('baz')

    fjob = FeedJob()
    fjob.signal.register('output', qjob.enqueue)
    fjob.signal.register('finished', lambda fjob: qjob.close())
    qjob.start()
    fjob.start()

    await asyncio.gather(fjob.wait(), qjob.wait())

    assert qjob._enqueue_args == ()
    assert qjob.handled_inputs == ['foo', 'bar', 'baz']


@pytest.mark.asyncio
async def test_enqueue_method_when_job_is_closed(qjob, tmp_path):
    qjob.start()
    qjob.close()
    assert qjob.is_closed
    assert not qjob.is_finished
    with pytest.raises(RuntimeError, match=r'^Do not call enqueue\(\) after close\(\): foo$'):
        qjob.enqueue('check this out')


@pytest.mark.asyncio
async def test_enqueue_method_when_job_is_finished(qjob, tmp_path):
    qjob.start()
    qjob.close()
    await qjob.wait()
    assert qjob.is_closed
    assert qjob.is_finished
    with pytest.raises(RuntimeError, match=r'^Do not call enqueue if job is already finished: foo$'):
        qjob.enqueue('check this out')


@pytest.mark.asyncio
async def test_handle_input_raises_exception(tmp_path, mocker):
    class FeedJob(JobBase):
        name = 'feeder'
        label = 'Feeder'
        cache_directory = str(tmp_path)

        async def run(self):
            self.send('foo')
            self.send('bar')
            self.send('baz')

    class QueueJob(FooJob):
        def handle_input(self, input):
            if 'b' in input:
                raise ValueError('No b allowed!')
            else:
                return super().handle_input(input)

    fjob = FeedJob()
    qjob = QueueJob()
    fjob.signal.register('output', qjob.enqueue)
    fjob.signal.register('finished', lambda fjob: qjob.close())
    qjob.start()
    fjob.start()

    await fjob.wait()
    with pytest.raises(ValueError, match=r'^No b allowed!$'):
        await qjob.wait()

    assert qjob._enqueue_args == ()
    assert qjob.handled_inputs == ['foo']


def test_error(qjob, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.jobs.base.JobBase.error'),
        'JobBase_error',
    )
    mocks.attach_mock(
        mocker.patch.object(qjob, 'close'),
        'close',
    )

    qjob.error('Everything is terrible.')

    assert mocks.mock_calls == [
        call.close(),
        call.JobBase_error('Everything is terrible.'),
    ]



import asyncio
from unittest.mock import Mock

import pytest


# Platform independent fs.filename_sanitization() mocking fixture from
# tests/conftest.py
@pytest.fixture(scope='module', autouse=True)
def strict_filename_sanitization(strict_filename_sanitization):
    pass


@pytest.fixture(scope='module', autouse=True)
def disable_http_requests(pytestconfig, module_mocker):
    if not pytestconfig.getoption('--allow-requests', None):
        # We sleep for a small amount of time to allow cached resonses to be
        # read and returned before concurrently running requests raise the
        # RuntimeError. This is important for requests made in predbs.multi.
        async def raise_after_delay(*_, **__):
            await asyncio.sleep(1)
            raise RuntimeError('HTTP requests are disabled; use --allow-requests')

        # We can't patch utils.http._request() because we want it to return
        # cached requests. utils.http._request() only uses
        # httpx.AsyncClient.send() so we can patch that.
        module_mocker.patch('httpx.AsyncClient.send', Mock(side_effect=raise_after_delay))



import importlib
import re
from unittest.mock import patch

import pytest

from upsies.trackers.base import exclude


# Run each test with POSIX and Windows path separator.
@pytest.fixture(params=('/', '\\'))
def sep(request, mocker):
    with patch('os.sep', request.param):
        # Reload module with patched os.sep.
        importlib.reload(exclude)

        def convert_path(path):
            # Convert POSIX path to Windows path.
            if exclude.sep == re.escape('\\'):
                return 'X:\\' + path.replace('/', '\\')
            else:
                return path

        yield convert_path

    # Reload exclude module again with the original os.sep.
    importlib.reload(exclude)


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        # Paths that must match
        ('/path/to/Foo.2012.BluRay.x264-ASDF.sfv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.SFV', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.Sfv', True),
        # Paths that must not match
        ('/path/to/sfv.2012.bluray.x264-asdf.mkv', False),
    ),
)
def test_checksums(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.checksums)
    match = regex.search(path)
    assert bool(match) is exp_match


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        # Paths that must match
        ('/path/to/Foo.S01.BluRay.x264-ASDF/Foo.S01.Extras.DVDRip.x264-ASDF', True),
        ('/path/to/Foo.S02.Extras.BluRay.x264-ASDF', True),
        ('/path/to/Foo.S03.720p.BluRay.x264-ASDF/Extras/Foo.S03.Extras.S03.720p.BluRay.DD2.0.x264-ASDF.mkv', True),
        ('/path/to/Foo.S04.720p.BluRay.x264-ASDF/Extras/Foo.S04.Extras.Topic.S03.720p.BluRay.DD2.0.x264-ASDF.mkv', True),
        ('/path/to/Foo.S05.720p.BluRay.x264-ASDF/Foo.S05.Extra.1.720p.BluRay.x264-EbP.mkv', True),

        # Paths that must not match
        ('/path/to/Extras.S01.720p.BluRay.x264-ASDF', False),
        ('/path/to/Foo.S06.720p.BluRay.x264-ASDF/Foo.S06E01.Extraordinary.720p.BluRay-ASDF.mkv', False),
        ('/path/to/Foo.S07.720p.BluRay.x264-ASDF/Foo.S07E02.Extra.Ordinary.720p.BluRay-ASDF.mkv', False),
        ('/path/to/the.extractors.2019.720p.bluray.x264-brmp.mkv', False),
    ),
)
def test_extras(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.extras)
    match = regex.search(path)
    assert bool(match) is exp_match


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        ('/path/to/Foo.2012.BluRay.x264-ASDF.png', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.PNG', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.jpg', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.Jpg', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.jpeg', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.jPEG', True),
    ),
)
def test_images(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.images)
    match = regex.search(path)
    assert bool(match) is exp_match


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        ('/path/to/Foo.2012.BluRay.x264-ASDF.nfo', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.NFO', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.Nfo', True),
    ),
)
def test_nfo(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.nfo)
    match = regex.search(path)
    assert bool(match) is exp_match


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        # Paths that must match
        ('/path/to/Foo.2012.BluRay.x264-ASDF/!Sample/Foo.2012.BluRay.x264-ASDF.sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/!sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/Foo.2012.BluRay.x264-ASDF sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/SAMPLE-Foo.2012.BluRay.x264-ASDF.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/Sample/Foo.2012.BluRay.x264-ASDF.sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/Sample/f-foo-x264-sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/asdf-foo.sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/foo.2012.bluray.x264.sample-asdf.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/sample.mkv', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF/sample/Foo.2012.BluRay.x264-ASDF.sample.mkv', True),
        # Paths that must not match
        ('/path/to/Sample.2012.BluRay.x264-ASDF.mkv', False),
        ('/path/to/Sample.2012.BluRay.x264-ASDF/Sample.2012.BluRay.x264-ASDF.mkv', False),
        ('/path/to/The.Sample.2012.BluRay.x264-ASDF.mkv', False),
        ('/path/to/The.Sample.2012.BluRay.x264-ASDF/The.Sample.2012.BluRay.x264-ASDF.mkv', False),
    ),
)
def test_samples(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.samples)
    match = regex.search(path)
    assert bool(match) is exp_match


@pytest.mark.parametrize(
    argnames='path, exp_match',
    argvalues=(
        ('/path/to/Foo.2012.BluRay.x264-ASDF.srt', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.SRT', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.Srt', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.sub', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.SUB', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.sUb', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.idx', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.IDX', True),
        ('/path/to/Foo.2012.BluRay.x264-ASDF.idX', True),
    ),
)
def test_subtitles(path, exp_match, sep):
    path = sep(path)
    regex = re.compile(exclude.subtitles)
    match = regex.search(path)
    assert bool(match) is exp_match



import datetime
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import bs4
import pytest

from upsies import __project_name__, errors
from upsies.trackers.uhd import UhdTracker, UhdTrackerConfig, UhdTrackerJobs


@pytest.fixture
def make_tracker():
    def make_tracker(**kwargs):
        options = {
            'username': 'bunny',
            'password': 'hunter2',
            'base_url': 'http://uhd.local',
        }
        options.update(kwargs)
        return UhdTracker(options=options)

    return make_tracker


@pytest.fixture
def tracker(make_tracker):
    return make_tracker()


def test_name_attribute():
    assert UhdTracker.name == 'uhd'


def test_label_attribute():
    assert UhdTracker.label == 'UHD'


def test_TrackerConfig_attribute():
    assert UhdTracker.TrackerConfig is UhdTrackerConfig


def test_TrackerJobs_attribute():
    assert UhdTracker.TrackerJobs is UhdTrackerJobs


def test_base_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._base_url == 'http://foo.local'


def test_login_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._login_url == 'http://foo.local/login.php'


def test_logout_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._logout_url == 'http://foo.local/logout.php'


def test_ajax_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._ajax_url == 'http://foo.local/ajax.php'


def test_upload_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._upload_url == 'http://foo.local/upload.php'


def test_torrents_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._torrents_url == 'http://foo.local/torrents.php'


@pytest.mark.parametrize('method', ('POST', 'Get'))
@pytest.mark.parametrize(
    argnames='exception, error_prefix, exp_exception',
    argvalues=(
        (None, 'Oh no', None),
        (errors.RequestError('Upload failed'), 'Oh no', errors.RequestError('Oh no: Upload failed')),
        (errors.RequestError('Upload failed'), '', errors.RequestError('Upload failed')),
        (RuntimeError('Everything failed'), 'OMG', RuntimeError('Everything failed')),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__request(
        method,
        exception, error_prefix, exp_exception,
        tracker, mocker,
):
    kwargs = {'foo': 'bar'}

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.http.get', AsyncMock(return_value='GET response')), 'get')
    mocks.attach_mock(mocker.patch('upsies.utils.http.post', AsyncMock(return_value='POST response')), 'post')

    if exp_exception:
        getattr(mocks, method.lower()).side_effect = exception
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await tracker._request(method, 'asdf', error_prefix=error_prefix, **kwargs)
    else:
        return_value = await tracker._request(method, 'asdf', error_prefix=error_prefix, **kwargs)
        assert return_value == f'{method.upper()} response'

    assert mocks.mock_calls == [
        getattr(call, method.lower())('asdf', user_agent=True, cache=False, foo='bar')
    ]


@pytest.mark.asyncio
async def test__failed_to_find_error(tracker, mocker):
    msg_prefix = 'WTF'
    doc = Mock()
    exp_filepath = 'WTF.uhd.html'
    exp_msg = (
        f'{msg_prefix}: '
        f'No error message found (dumped HTML response to {exp_filepath})'
    )

    html_dump_mock = mocker.patch('upsies.utils.html.dump')

    with pytest.raises(RuntimeError, match=rf'^{re.escape(exp_msg)}$'):
        tracker._failed_to_find_error(doc, msg_prefix)

    html_dump_mock.call_args_list == [call(doc, exp_filepath)]


@pytest.fixture
def mocks_for_login(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_request'), '_request')
    mocks.attach_mock(mocker.patch.object(tracker, '_confirm_login'), '_confirm_login')
    return mocks

@pytest.mark.asyncio
async def test_login_without_username(mocks_for_login, tracker):
    del tracker.options['username']
    with pytest.raises(errors.RequestError, match=r'^Login failed: No username configured$'):
        await tracker._login()
    assert mocks_for_login.mock_calls == []

@pytest.mark.asyncio
async def test_login_without_password(mocks_for_login, tracker):
    del tracker.options['password']
    with pytest.raises(errors.RequestError, match=r'^Login failed: No password configured$'):
        await tracker._login()
    assert mocks_for_login.mock_calls == []

@pytest.mark.asyncio
async def test_login_succeeds(mocks_for_login, tracker):
    await tracker._login()
    assert mocks_for_login.mock_calls == [
        call._request(
            method='POST',
            url=tracker._login_url,
            data={
                'username': tracker.options['username'],
                'password': tracker.options['password'],
                'two_step': '',  # 2FA
                'login': 'Log in',
            },
            error_prefix='Login failed',
        ),
        call._confirm_login(mocks_for_login._request.return_value),
    ]


@pytest.mark.parametrize(
    argnames='response, exp_result, exp_failed_to_find_error_called, exp_auth',
    argvalues=(
        (
            '<a href="logout.php?foo=bar&auth=d34db33f">Logout</a>',
            None,
            False,
            'd34db33f',
        ),
        (
            '<html></html>',
            RuntimeError('Login failed: No error message found'),
            True,
            None,
        ),
        (
            (
                '<form action="login.php">'
                '<span class="warning"><br/>Authentication failed<br /><br /></span>'
                '<br/>You have <span class="info">5</span> attempts remaining.<br /><br />'
                '<strong>WARNING:</strong> You will be banned for 6 hours after your attempts run out!<br /><br />'
                '<table class="layout">'
                'IGNORED MORE TEXT'
                '</table>'
            ),
            errors.RequestError(
                'Login failed: Authentication failed\n'
                'You have 5 attempts remaining.\n'
                'WARNING: You will be banned for 6 hours after your attempts run out!'
            ),
            False,
            None,
        ),
    ),
    ids=lambda v: repr(v),
)
def test__confirm_login(response, exp_result, exp_failed_to_find_error_called, exp_auth, tracker, mocker):
    mocker.patch.object(tracker, '_failed_to_find_error')
    exp_doc = bs4.BeautifulSoup(response, features='html.parser')

    if isinstance(exp_result, Exception):
        tracker._failed_to_find_error.side_effect = exp_result
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker._confirm_login(response)
    else:
        return_value = tracker._confirm_login(response)
        assert return_value == exp_result

    if exp_failed_to_find_error_called:
        assert tracker._failed_to_find_error.call_args_list == [
            call(exp_doc, 'Login failed'),
        ]
    else:
        assert tracker._failed_to_find_error.call_args_list == []

    if exp_auth is None:
        assert not hasattr(tracker, '_auth')
    else:
        assert tracker._auth == exp_auth


@pytest.mark.parametrize(
    argnames='request_exception',
    argvalues=(
        None,
        errors.RequestError('Connection refused'),
        RuntimeError('Connection exploded'),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__logout(request_exception, tracker, mocker):
    mocker.patch.object(tracker, '_request', side_effect=request_exception)
    auth = tracker._auth = 'd34db33f'

    if request_exception:
        with pytest.raises(type(request_exception), match=rf'^{re.escape(str(request_exception))}$'):
            await tracker._logout()
    else:
        await tracker._logout()

    assert tracker._request.call_args_list == [call(
        method='GET',
        url=tracker._logout_url,
        params={'auth': auth},
        error_prefix='Logout failed',
    )]
    assert not hasattr(tracker, '_auth')


@pytest.mark.asyncio
async def test_get_announce_url_from_config(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
    mocks.attach_mock(mocker.patch.object(tracker, '_request'), '_request')
    mocker.patch.object(type(tracker), 'options', PropertyMock(return_value={
        'announce_url': 'http://foo.bar/announce',
    }))

    announce_url = await tracker.get_announce_url()
    assert announce_url == 'http://foo.bar/announce'
    assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames='html, exp_result',
    argvalues=(
        (
            '<html><input value="http://foo.local:123/announce" /></html>',
            'http://foo.local:123/announce',
        ),
        (
            '<html><input value="https://foo.local:123/announce?d34db33f" /></html>',
            'https://foo.local:123/announce?d34db33f',
        ),
        (
            '<html><input value="https://foo.local:123/announce/d34db33f" /></html>',
            'https://foo.local:123/announce/d34db33f',
        ),
        (
            '<html><input value="https://foo.local:123/d34db33f/announce" /></html>',
            'https://foo.local:123/d34db33f/announce',
        ),
        (
            '<html><input value="https://foo.local:123/anon" /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.uhd.announce_url YOUR_URL',
            ),
        ),
        (
            '<html><input value="" /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.uhd.announce_url YOUR_URL',
            ),
        ),
        (
            '<html><input /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.uhd.announce_url YOUR_URL',
            ),
        ),
        (
            '<html><div>hello</div></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.uhd.announce_url YOUR_URL',
            ),
        ),
    ),
)
@pytest.mark.asyncio
async def test_get_announce_url_from_website(html, exp_result, tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
    mocks.attach_mock(mocker.patch.object(tracker, '_request', return_value=html), '_request')
    tracker.options.pop('announce_url', None)

    if isinstance(exp_result, BaseException):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await tracker.get_announce_url()
    else:
        return_value = await tracker.get_announce_url()
        assert return_value == exp_result

    assert mocks.mock_calls == [
        call.login(),
        call._request(method='GET', url=tracker._upload_url),
    ]


@pytest.fixture
def mocks_for_get_uhd_info(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_get_uhd_info'), '_get_uhd_info')
    mocks.attach_mock(mocker.patch('asyncio.sleep'), 'sleep')
    return mocks

@pytest.mark.asyncio
async def test_get_uhd_info_with_unknown_imdb_id(mocks_for_get_uhd_info, tracker):
    with pytest.raises(AssertionError, match=r'^IMDb ID is not available yet$'):
        await tracker.get_uhd_info(None)
    assert mocks_for_get_uhd_info.mock_calls == []

@pytest.mark.asyncio
async def test_get_uhd_info_with_readily_available_info(mocks_for_get_uhd_info, tracker):
    imdb_id = 'tt123456'
    mocks_for_get_uhd_info._get_uhd_info.return_value = {'error': '0', 'the': 'info'}

    return_value = await tracker.get_uhd_info(imdb_id)
    assert return_value == {'error': '0', 'the': 'info'}
    assert mocks_for_get_uhd_info.mock_calls == [
        call._get_uhd_info(imdb_id),
    ]

@pytest.mark.parametrize(
    argnames='imdb_id, responses, exp_mock_calls',
    argvalues=(
        (
            'tt123456',
            (
                {'error': '1', 'message': 'Fetching'},
                {'error': '0', 'the': 'info'},
            ),
            [
                call._get_uhd_info('tt123456'),
                call.sleep(6),
                call._get_uhd_info('tt123456'),
            ],
        ),
        (
            'tt123456',
            (
                {'error': '1', 'message': 'Fetching'},
                {'error': '1', 'message': 'Loading'},
                {'error': '0', 'the': 'info'},
            ),
            [
                call._get_uhd_info('tt123456'),
                call.sleep(6),
                call._get_uhd_info('tt123456'),
                call.sleep(5),
                call._get_uhd_info('tt123456'),
            ],
        ),
        (
            'tt123456',
            (
                {'error': '1', 'message': 'Fetching'},
                {'error': '1', 'message': 'Loading'},
                {'error': '1', 'message': 'Still loading'},
                {'error': '0', 'the': 'info'},
            ),
            [
                call._get_uhd_info('tt123456'),
                call.sleep(6),
                call._get_uhd_info('tt123456'),
                call.sleep(5),
                call._get_uhd_info('tt123456'),
                call.sleep(4),
                call._get_uhd_info('tt123456'),
            ],
        ),
        (
            'tt123456',
            (
                {'error': '1', 'message': 'Fetching'},
                {'error': '1', 'message': 'Loading'},
                {'error': '1', 'message': 'Still loading'},
                {'error': '1', 'message': '...'},
                {'error': '0', 'the': 'info'},
            ),
            [
                call._get_uhd_info('tt123456'),
                call.sleep(6),
                call._get_uhd_info('tt123456'),
                call.sleep(5),
                call._get_uhd_info('tt123456'),
                call.sleep(4),
                call._get_uhd_info('tt123456'),
                call.sleep(3),
                call._get_uhd_info('tt123456'),
            ],
        ),
        (
            'tt123456',
            (
                {'error': '1', 'message': 'Fetching'},
                {'error': '1', 'message': 'Loading'},
                {'error': '1', 'message': 'Still loading'},
                {'error': '1', 'message': '...'},
                {'error': '1', 'message': '!!!'},
                {'error': '0', 'the': 'info'},
            ),
            [
                call._get_uhd_info('tt123456'),
                call.sleep(6),
                call._get_uhd_info('tt123456'),
                call.sleep(5),
                call._get_uhd_info('tt123456'),
                call.sleep(4),
                call._get_uhd_info('tt123456'),
                call.sleep(3),
                call._get_uhd_info('tt123456'),
                call.sleep(3),
                call._get_uhd_info('tt123456'),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_get_uhd_info_with_info_getting_fetched_by_server(
        imdb_id, responses, exp_mock_calls,
        mocks_for_get_uhd_info, tracker,
):
    mocks_for_get_uhd_info._get_uhd_info.side_effect = responses

    return_value = await tracker.get_uhd_info(imdb_id)
    assert return_value == responses[-1]
    assert mocks_for_get_uhd_info.mock_calls == exp_mock_calls

@pytest.mark.asyncio
async def test_get_uhd_info_with_server_fetching_too_long(mocks_for_get_uhd_info, tracker):
    imdb_id = 'tt123456'
    mocks_for_get_uhd_info._get_uhd_info.side_effect = (
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
        {'error': '1', 'message': 'Fetching'},
    )

    return_value = await tracker.get_uhd_info(imdb_id)
    assert return_value == {}
    assert mocks_for_get_uhd_info.mock_calls == [
        call._get_uhd_info(imdb_id),
        call.sleep(6),
        call._get_uhd_info(imdb_id),
        call.sleep(5),
        call._get_uhd_info(imdb_id),
        call.sleep(4),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
        call.sleep(3),
        call._get_uhd_info(imdb_id),
    ]

@pytest.mark.asyncio
async def test_get_uhd_info_with_server_request_error(mocks_for_get_uhd_info, tracker):
    imdb_id = 'tt123456'
    mocks_for_get_uhd_info._get_uhd_info.side_effect = (
        {'error': '1', 'message': 'Fetching'},
        errors.RequestError('Connection refused'),
    )

    return_value = await tracker.get_uhd_info(imdb_id)
    assert return_value == {}
    assert mocks_for_get_uhd_info.mock_calls == [
        call._get_uhd_info(imdb_id),
        call.sleep(6),
        call._get_uhd_info(imdb_id),
    ]


@pytest.mark.asyncio
async def test__get_uhd_info(tracker, mocker):
    imdb_id = 'tt123456'
    mocks = Mock()
    mocks.response.json.return_value = {'the': 'info'}
    mocks.now.return_value = datetime.datetime(2012, 11, 10, 9, 8, 7, 654321)
    mocker.patch('datetime.datetime', Mock(now=mocks.now))
    mocks.attach_mock(mocker.patch.object(tracker, '_request', return_value=mocks.response), '_request')

    return_value = await tracker._get_uhd_info(imdb_id)
    assert return_value == {'the': 'info'}
    assert mocks.mock_calls == [
        call.now(datetime.timezone.utc),
        call._request(
            method='GET',
            url=tracker._ajax_url,
            params={
                'action': 'imdb_fetch',
                'imdbid': imdb_id,
                '_': int(mocks.now.return_value.timestamp() * 1000),
            },
        ),
        call.response.json(),
    ]


@pytest.fixture
def mocks_for_upload(tracker, mocker):
    tracker_jobs = Mock(
        post_data={'post': 'data'},
        post_files={'post': 'files'},
    )
    mocker.patch.object(type(tracker), 'is_logged_in', PropertyMock(return_value=True))
    mocker.patch.object(type(tracker), '_auth', PropertyMock(return_value='d34db33f'), create=True)
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_request'), '_request')
    mocks.attach_mock(mocker.patch.object(tracker, '_handle_upload_response'), '_handle_upload_response')
    return mocks, tracker_jobs

@pytest.mark.asyncio
async def test_upload_if_not_logged_in(mocks_for_upload, tracker, mocker):
    mocks, tracker_jobs = mocks_for_upload
    mocker.patch.object(type(tracker), 'is_logged_in', PropertyMock(return_value=False))
    with pytest.raises(AssertionError):
        await tracker.upload(tracker_jobs)
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_upload_if_logged_in(mocks_for_upload, tracker):
    mocks, tracker_jobs = mocks_for_upload
    await tracker.upload(tracker_jobs)
    assert mocks.mock_calls == [
        call._request(
            method='POST',
            url=tracker._upload_url,
            data=tracker_jobs.post_data,
            files=tracker_jobs.post_files,
            follow_redirects=False,
        ),
        call._handle_upload_response(mocks._request.return_value),
    ]


class MockResponse(str):
    def __new__(cls, *args, headers={}, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self.headers = headers
        return self

@pytest.fixture
def mocks_for__handle_post_response(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_failed_to_find_error'), '_failed_to_find_error')
    return mocks

@pytest.mark.parametrize('parameters_prefix', ('torrents.php', '{tracker._torrents_url}'))
def test__handle_upload_response_with_redirect_response_to_torrent_url(
        parameters_prefix,
        mocks_for__handle_post_response, tracker, mocker,
):
    response = MockResponse(
        headers={'Location': parameters_prefix.format(tracker=tracker) + '?id=123456'}
    )
    return_value = tracker._handle_upload_response(response)
    assert return_value == tracker._torrents_url + '?id=123456'
    assert mocks_for__handle_post_response.mock_calls == []

def test__handle_upload_response_with_redirect_response_to_nontorrent_url(mocks_for__handle_post_response, tracker, mocker):
    response = MockResponse(
        '<html></html>',
        headers={'Location': 'https://evil.org' + '?id=123456'}
    )
    return_value = tracker._handle_upload_response(response)
    assert return_value is None
    assert mocks_for__handle_post_response.mock_calls == [
        call._failed_to_find_error(bs4.BeautifulSoup(response, features='html.parser'), 'Upload failed'),
    ]

def test__handle_upload_response_without_redirect_response_and_error_message_found(
        mocks_for__handle_post_response, tracker, mocker,
):
    response = MockResponse(
        '<html>'
        '  <div id="scontent">'
        '    <div class="thin">'
        '      <p>Your personal announce URL is ...</p>'
        '      <p>This is <b>the</b> error message!</p>'
        '    </div>'
        '  </div>'
        '</html>'
    )
    exp_msg = 'Upload failed: This is the error message!'
    with pytest.raises(errors.RequestError, match=rf'^{re.escape(exp_msg)}$'):
        tracker._handle_upload_response(response)
    assert mocks_for__handle_post_response.mock_calls == []

def test__handle_upload_response_without_redirect_response_and_no_error_message_found(
        mocks_for__handle_post_response, tracker, mocker,
):
    response = MockResponse(
        '<html>'
        '  <div id="scontent">'
        '    <div class="wide">'
        '      Something that is not an error message.'
        '    </div>'
        '  </div>'
        '</html>'
    )
    tracker._handle_upload_response(response)
    assert mocks_for__handle_post_response.mock_calls == [
        call._failed_to_find_error(bs4.BeautifulSoup(response, features='html.parser'), 'Upload failed'),
    ]



import base64

import pytest

from upsies import utils
from upsies.trackers.uhd import UhdTrackerConfig


@pytest.fixture
def tracker_config():
    return UhdTrackerConfig()


def test_UhdTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'base_url',
        'username',
        'password',
        'anonymous',
        'announce_url',
        'source',
        'image_host',
        'screenshots',
        'exclude',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
        'randomize_infohash',
    }


def test_UhdTrackerConfig_defaults_base_url(tracker_config):
    assert tracker_config['base_url'] == base64.b64decode('aHR0cHM6Ly91aGRiaXRzLm9yZw==').decode('ascii')


def test_UhdTrackerConfig_defaults_username(tracker_config):
    assert tracker_config['username'] == ''


def test_UhdTrackerConfig_defaults_password(tracker_config):
    assert tracker_config['password'] == ''


def test_UhdTrackerConfig_defaults_anonymous(tracker_config):
    assert isinstance(tracker_config['anonymous'], utils.types.Bool)
    assert not tracker_config['anonymous']
    assert tracker_config['anonymous'] == 'no'
    assert tracker_config['anonymous'].description == (
        'Whether your username is displayed on your uploads.'
    )


def test_UhdTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == ''
    assert tracker_config['announce_url'].description == (
        'Your personal announce URL.'
    )


def test_UhdTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == '[UHDBits]'


def test_UhdTrackerConfig_defaults_image_host(tracker_config, assert_config_list_of_choice):
    exp_options = utils.imghosts.imghost_names()
    assert_config_list_of_choice(
        items=tracker_config['image_host'],
        exp_items=('ptpimg', 'freeimage', 'imgbox'),
        exp_options=exp_options,
        exp_description=(
            'List of image hosting service names. The first service is normally used '
            + 'with the others as backup if uploading to the first fails.\n'
            + 'Supported services: ' + ', '.join(exp_options)
        ),
    )


def test_UhdTrackerConfig_defaults_screenshots(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots'],
        value=4,
        min=2,
        max=10,
        description='How many screenshots to make.',
    )


def test_UhdTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == ()


def test_UhdTrackerConfig_arguments(tracker_config):
    exp_argument_definitions = {
        'submit': {
            ('--anonymous', '--an'),
            ('--internal', '--in'),
            ('--3d',),
            ('--vie', '--vi'),
            ('--screenshots', '--ss'),
            ('--poster', '--po'),
            ('--trailer', '--tr'),
            ('--only-description', '--od'),
        },
    }
    assert set(tracker_config.argument_definitions) == set(exp_argument_definitions)
    for command in exp_argument_definitions:
        assert set(tracker_config.argument_definitions[command]) == exp_argument_definitions[command]


def test_UhdTrackerConfig_argument_definitions_submit_anonymous(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--anonymous', '--an')] == {
        'help': 'Hide your username for this submission',
        'action': 'store_true',
    }


def test_UhdTrackerConfig_argument_definitions_submit_internal(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--internal', '--in')] == {
        'help': 'Internal encode (use only if you were told to)',
        'action': 'store_true',
    }


def test_UhdTrackerConfig_argument_definitions_submit_3d(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--3d',)] == {
        'help': 'Mark this as a 3D release',
        'action': 'store_true',
    }


def test_UhdTrackerConfig_argument_definitions_submit_vie(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--vie', '--vi')] == {
        'help': 'Release contains Vietnamese audio dub',
        'action': 'store_true',
    }


def test_UhdTrackerConfig_argument_definitions_submit_screenshots(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--screenshots', '--ss')] == {
        'help': 'How many screenshots to make',
        'type': utils.argtypes.number_of_screenshots(min=2, max=10),
    }


def test_UhdTrackerConfig_argument_definitions_submit_poster(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--poster', '--po')] == {
        'help': 'Path or URL to poster image (autodetected by default)',
    }


def test_UhdTrackerConfig_argument_definitions_submit_trailer(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--trailer', '--tr')] == {
        'help': 'Trailer YouTube ID or URL (autodetected by default)',
    }


def test_UhdTrackerConfig_argument_definitions_submit_only_descrption(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--only-description', '--od')] == {
        'help': 'Only generate description (do not upload anything)',
        'action': 'store_true',
        'group': 'generate-metadata',
    }



"""
Query online databases like IMDb
"""

import asyncio
import collections

from .. import errors
from ..utils import webdbs
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class WebDbSearchJob(JobBase):
    """
    Ask user to select a specific search result from an internet database

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``search_results``
            Emitted after new search results are available. Registered callbacks
            get a sequence of :class:`~.utils.webdbs.common.SearchResult`
            instances as a positional argument.

        ``searching_status``
            Emitted when a new search is started and when new results are
            available. Registered callbacks get `True` for "search started" or
            `False` for "search ended" as a positional argument.

        ``info_updating``
            Emitted when an attempt to fetch additional information is made.
            Registered callbacks get the name of an attribute of a
            :class:`~.utils.webdbs.common.SearchResult` object that returns a
            coroutine function as a positional argument.

        ``info_updated``
            Emitted when additional information is available. Registered
            callbacks are called for each piece of information and get ``key``
            and ``value`` as positional arguments. ``key`` is an attribute of a
            :class:`~.utils.webdbs.common.SearchResult` object that returns a
            coroutine function. ``value`` is the return value of that coroutine
            function.

        ``query_updated``
            Emitted after the :attr:`query` changed. Registered callbacks get
            the new query as a positional argument.
    """

    @property
    def name(self):
        return f'{self._db.name}-id'

    @property
    def label(self):
        return f'{self._db.label} ID'

    @property
    def cache_id(self):
        """Database :attr:`~.WebDbApiBase.name` and :attr:`query`"""
        return self.query

    @property
    def db(self):
        """:class:`~.webdbs.base.WebDbApiBase` instance"""
        return self._db

    @property
    def query(self):
        """:class:`~.webdbs.common.Query` instance"""
        return self._query

    @property
    def no_id_ok(self):
        """Whether this job will finish gracefully if no ID is selected"""
        return self._no_id_ok

    @no_id_ok.setter
    def no_id_ok(self, value):
        self._no_id_ok = bool(value)

    @property
    def is_searching(self):
        """Whether a search request is currently being made"""
        return self._is_searching

    def initialize(self, *, db, query, no_id_ok=False):
        """
        Set internal state

        :param WebDbApiBase db: Return value of :func:`.utils.webdbs.webdb`
        :param query: Path of the release (doesn't have to exist) or
            :class:`~.webdbs.common.Query` instance
        :param no_id_ok: Whether this job may finish without error if no ID is
            selected
        """
        assert isinstance(db, webdbs.WebDbApiBase), f'Not a WebDbApiBase: {db!r}'
        self._db = db
        if not isinstance(query, webdbs.Query):
            query = webdbs.Query.from_any(query)
        self._query = self._db.sanitize_query(query)

        self.signal.add('search_results')
        self.signal.add('searching_status')
        self.signal.add('info_updating')
        self.signal.add('info_updated')
        self.signal.add('query_updated')
        self.signal.add('selected', record=True)

        # Call search() when `self.query` is changed externally, e.g. by another
        # job. This happens if we have multiple WebDbSearchJobs with different
        # WebDbs and we want to update the query of the second job with accurate
        # information from the first job. This means the user only has to fix a
        # potentially badly autodetected query once and not for each
        # WebDbSearchJob.
        self._query.signal.register('changed', self.search)

        # Hold some information about the selected search result.
        self._selected = {}
        self.signal.register('selected', self._selected.update)

        self.no_id_ok = no_id_ok
        self._is_searching = False
        self._search_task = None
        self._info_callbacks_task = None
        self._info_callbacks = _InfoCallbacks(
            callbacks={
                'id': self._make_update_info_func('id'),
                'summary': self._make_update_info_func('summary'),
                'title_original': self._make_update_info_func('title_original'),
                'title_english': self._make_update_info_func('title_english'),
                'genres': self._make_update_info_func('genres'),
                'directors': self._make_update_info_func('directors'),
                'cast': self._make_update_info_func('cast'),
                'countries': self._make_update_info_func('countries'),
            },
            error_callback=self.warn,
        )

    def _make_update_info_func(self, key):
        return lambda value: self._update_info_value(key, value)

    def _update_info_value(self, attr, value):
        if value is Ellipsis:
            # Indicate that we are updating `attr` by not providing a value.
            self.signal.emit('info_updating', attr)
        else:
            self.signal.emit('info_updated', attr, value)

    async def run(self):
        # Trigger initial search.
        self.search(self.query)

        # Block until finalize() is called.
        await self.finalization()

    async def _search(self, query):
        self._set_state(is_searching=True, results=())

        try:
            # Get new search results
            results = await self._db.search(query)
            self._set_state(is_searching=False, results=results)

            # Get details about the first search result
            self._run_info_callbacks(results[0] if results else None)

        except errors.RequestError as e:
            self.warn(e)

        finally:
            self._set_state(is_searching=False)

    def _set_state(self, *, is_searching=None, results=None):
        if is_searching is not None:
            self._is_searching = bool(is_searching)
            self.signal.emit('searching_status', self._is_searching)
        if results is not None:
            self.signal.emit('search_results', results)

    def _run_info_callbacks(self, result):
        if self._info_callbacks_task:
            self._info_callbacks_task.cancel()
        self._info_callbacks_task = self.add_task(
            self._info_callbacks(result),
            # Auto-remove reference to task when it is done.
            callback=self._unset_info_callbacks_task,
        )

    def _unset_info_callbacks_task(self, task_):
        self._info_callbacks_task = None

    def _cancel_tasks(self):
        for task_attr in ('_search_task', '_info_callbacks_task'):
            task = getattr(self, task_attr, None)
            if task:
                task.cancel()
                setattr(self, task_attr, None)

    def search(self, query):
        """
        Make new search request

        Any currently ongoing search is cancelled.

        :param str query: Query that is first passed through
            :meth:`~.webdbs.Query.from_any` and then
            :meth:`~.webdbs.WebDbApiBase.sanitize_query`
        """
        # Stop any ongoing requests. They are outdated.
        self._cancel_tasks()

        # Remove any warnings from previous requests (e.g. network issues).
        self.clear_warnings()

        # Update search query silently (without emitting any signals).
        new_query = self._db.sanitize_query(webdbs.Query.from_any(query))
        self.query.update(new_query, silent=True)
        self.signal.emit('query_updated', self.query)

        # This method is called to fix the initial query. If this happens before
        # this job is started, we cannot add any tasks yet. But that's ok
        # because we already updated our internal query, which will be used for
        # the initial search if this job is started.
        if self.was_started:
            # Initiate new search in a separate task so we can cancel it.
            self._search_task = self.add_task(
                self._search(new_query),
                # Auto-remove reference to task when it is done.
                callback=self._unset_search_task,
            )

    def _unset_search_task(self, task_):
        self._search_task = None

    def result_focused(self, result):
        """
        Must be called by the UI when the user focuses a different search result

        :param SearchResult result: Focused search result or `None` if there are
            no results
        """
        self._run_info_callbacks(result)

    def result_selected(self, result):
        """
        Must be called by the UI when the user accepts a search result

        :param SearchResult result: Focused search result or `None` if there are
            no results and the user accepted that (see :attr:`no_id_ok`)
        """
        if not self.is_searching:
            if result is not None:
                self.signal.emit('selected', {
                    'id': result.id,
                    'title': result.title,
                    'type': result.type,
                    'url': result.url,
                    'year': result.year,
                })
                self.send(result.id)

            if result is not None or self.no_id_ok:
                self._cancel_tasks()
                self.finalize()

    @property
    def selected(self):
        """
        Some more information about the selected search result

        This is a :class:`dict` that is empty before :meth:`result_selected`
        is called. After :meth:`result_selected` is called, it contains the
        following keys:

            - ``id``
            - ``title``
            - ``type``
            - ``url``
            - ``year``
        """
        return self._selected.copy()

    @property
    def exit_code(self):
        """`0` if job was successful, `> 0` otherwise, None while job is not finished"""
        exit_code = super().exit_code
        if exit_code is not None:
            if exit_code != 0 and self.no_id_ok:
                return 0
            else:
                return exit_code


class _InfoCallbacks:
    def __init__(self, callbacks, error_callback):
        # `callbacks` maps names of SearchResult attributes to callbacks that
        # get the value of each attribute.
        #
        # For example, {"title": handle_title} means: Get the "title" attribute
        # from a search result and pass it to handle_title().
        #
        # Values of SearchResult attributes may also be coroutine functions that
        # return the value. In that case, call
        # handle_title(await search_result.title()).
        self._callbacks = callbacks
        self._error_callback = error_callback

    async def __call__(self, result):
        if not result:
            for callback in self._callbacks.values():
                callback('')
        else:
            self._update_text_values(result)
            await self._update_awaitable_values(result)

    def _update_text_values(self, result):
        # Update plain, non-callable values first before awaiting coroutines.
        for attr, callback in self._callbacks.items():
            value = getattr(result, attr)
            if not callable(value):
                callback(self._value_as_string(value))

    async def _update_awaitable_values(self, result):
        # Update values that are coroutine functions that return the value.
        tasks = []
        for attr, callback in self._callbacks.items():
            value = getattr(result, attr)
            if callable(value):
                tasks.append(self._call_callback(
                    value_getter=value,
                    callback=callback,
                    cache_key=(result.id, attr),
                ))
        # Get all values concurrently
        await asyncio.gather(*tasks)

    _cache = {}
    _delay_between_updates = 0.5

    async def _call_callback(self, value_getter, callback, cache_key):
        cached_value = self._cache.get(cache_key, None)
        if cached_value is not None:
            callback(cached_value)
        else:
            # Wait for a bit before starting to load info. If the user rapidly
            # focuses different search results, we don't want to start lots of
            # requests every few fractions of a second.
            callback(Ellipsis)
            await asyncio.sleep(self._delay_between_updates)

            # Get attribute value from from search result
            try:
                value = await value_getter()
            except errors.RequestError as e:
                callback('')
                self._error_callback(e)
            else:
                value_str = self._value_as_string(value)
                self._cache[cache_key] = value_str
                callback(value_str)

    @staticmethod
    def _value_as_string(value):
        if not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
            return ', '.join(str(v) for v in value)
        else:
            return str(value)



import asyncio
import sys
from unittest.mock import Mock, call

import pytest

from upsies.uis.tui import widgets


@pytest.mark.asyncio
async def test_ActivityIndicator_active_property(mocker):
    mocker.patch('upsies.uis.tui.widgets.ActivityIndicator._iterate')
    ai = widgets.ActivityIndicator()

    assert ai.active is False
    assert ai._iterate.call_args_list == []
    assert ai._iterate_task is None

    ai.active = 1
    assert ai.active is True
    assert ai._iterate.call_args_list == [call()]
    assert isinstance(ai._iterate_task, asyncio.Task)
    iterate_task = ai._iterate_task

    ai.active = 0
    assert ai.active is False
    assert ai._iterate.call_args_list == [call()]
    assert ai._iterate_task is None
    if sys.version_info >= (3, 11, 0):
        assert iterate_task.cancelling()
    ai.active = False
    assert ai.active is False
    assert ai._iterate.call_args_list == [call()]
    assert ai._iterate_task is None

    ai.active = 'yes'
    assert ai.active is True
    assert ai._iterate.call_args_list == [call(), call()]
    assert isinstance(ai._iterate_task, asyncio.Task)
    iterate_task = ai._iterate_task
    ai.active = True
    assert ai.active is True
    assert ai._iterate.call_args_list == [call(), call()]
    assert isinstance(ai._iterate_task, asyncio.Task)

    ai.active = ''
    assert ai.active is False
    assert ai._iterate.call_args_list == [call(), call()]
    assert ai._iterate_task is None
    if sys.version_info >= (3, 11, 0):
        assert iterate_task.cancelling()


@pytest.mark.asyncio
async def test_ActivityIndicator_enable_disable(mocker):
    ai = widgets.ActivityIndicator()
    mocker.patch.object(ai, '_iterate')
    assert ai.active is False
    ai.enable()
    assert ai.active is True
    ai.disable()
    assert ai.active is False


def test_ActivityIndicator_format_property(mocker):
    mocker.patch('upsies.uis.tui.widgets.ActivityIndicator._iterate')
    ai = widgets.ActivityIndicator()
    assert ai.format == '{indicator}'
    ai.format = '[{indicator}]'
    assert ai.format == '[{indicator}]'


def test_ActivityIndicator_text_property():
    ai = widgets.ActivityIndicator()
    mock_text = Mock()
    ai._text = mock_text
    assert ai.text is mock_text


@pytest.mark.asyncio
async def test_ActivityIndicator_iterate():

    def callback(text):
        print(text, exp_texts)
        if len(exp_texts) == 1:
            iterate_task.cancel()
        assert text == exp_texts.pop(0)

    callback = Mock(side_effect=callback)

    ai = widgets.ActivityIndicator(
        interval=0,
        format=':{indicator}:',
        states=('a', 'b', 'c'),
        callback=callback,
    )

    exp_texts = [':a:', ':b:', ':c:', ':a:', ':b:', ':c:']
    assert ai.text == exp_texts.pop(0)
    exp_callback_calls = [call.callback(txt) for txt in exp_texts]

    iterate_task = asyncio.get_running_loop().create_task(ai._iterate())
    with pytest.raises(asyncio.CancelledError):
        await iterate_task

    assert callback.call_args_list == exp_callback_calls


@pytest.mark.asyncio
async def test_ActivityIndicator_iterate_without_callback():
    ai = widgets.ActivityIndicator(
        interval=0,
        format='<{indicator}>',
        states=('a', 'b', 'c'),
    )

    exp_texts = ['<a>', '<b>', '<c>', '<a>', '<b>', '<c>']
    assert ai.text == exp_texts.pop(0)

    async def check_text_loop():
        while exp_texts:
            print(ai.text, exp_texts)
            assert ai.text == exp_texts.pop(0)
            # Yield control back to ActivityIndicator._iterate() so it can
            # change the text property to the next value.
            await asyncio.sleep(0)

        print('cancelling', iterate_task)
        iterate_task.cancel()

    iterate_task = asyncio.get_running_loop().create_task(ai._iterate())

    with pytest.raises(asyncio.CancelledError):
        await asyncio.gather(
            check_text_loop(),
            iterate_task,
        )


def test_ActivityIndicator___pt_container__():
    ai = widgets.ActivityIndicator()
    assert ai.__pt_container__() is ai.container
